use strict;
use File::Spec;
use File::Copy::Recursive qw(fcopy rcopy dircopy);

my $buildno = '0.1.2013.08.10';

print(STDERR <<"_END");
Assams $buildno
=======================================================================

This is a script to assemble many long sequences in parallel. Official
web site of this script is
http://www.fifthdimension.jp/products/assams/ .
To know script details, see above URL.

Copyright (C) 2011-2013  Akifumi S. Tanabe

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

_END

# display usage if command line options were not specified
unless (@ARGV) {
	&helpMessage();
}

# initialize variables
my $devnull = File::Spec->devnull();
my $numthreads = 1;
my @inputfiles;
my @qualfiles;
my $minident1 = 0.97;
my $minident2 = 0.97;
my $minident3 = 0.97;
my $minovllen = 100;
my $maxnhits = 10000;
my $strand;
my $alnwiggle = 11;
my $mergemode = 'normal';
my $noeid;
my $exact = 1;
my $ungapped;
my $gapped = 1;
my $outputbank = $ARGV[-1];
if (-e $outputbank) {
	&errorMessage(__LINE__, "\"$outputbank\" already exists.");
}
{
	my @temp = glob("$outputbank*");
	if (@temp) {
		&errorMessage(__LINE__, "Temporary file already exists.");
	}
}
my $minnseq = 1000;
my $maxnsplit;
# get arguments
{
	my %inputfiles;
	my %qualfiles;
	for (my $i = 0; $i < scalar(@ARGV) - 1; $i ++) {
		if ($ARGV[$i] =~ /^-+assemble(?:mode)?=(exact|ungapped|ungap|gapped|gap|e\+u\+g|e\+u|e\+g|u\+g|all|e|u|g|a)$/i) {
			if ($1 =~ /^(?:e|exact)$/i) {
				$exact = 1;
				$ungapped = 0;
				$gapped = 0;
			}
			elsif ($1 =~ /^(?:u|ungap|ungapped)$/i) {
				$exact = 0;
				$ungapped = 1;
				$gapped = 0;
			}
			elsif ($1 =~ /^(?:g|gap|gapped)$/i) {
				$exact = 0;
				$ungapped = 0;
				$gapped = 1;
			}
			elsif ($1 =~ /^e\+u$/i) {
				$exact = 1;
				$ungapped = 1;
				$gapped = 0;
			}
			elsif ($1 =~ /^e\+g$/i) {
				$exact = 1;
				$ungapped = 0;
				$gapped = 1;
			}
			elsif ($1 =~ /^u\+g$/i) {
				$exact = 0;
				$ungapped = 1;
				$gapped = 1;
			}
			elsif ($1 =~ /^(?:e\+u\+g|a|all)$/i) {
				$exact = 1;
				$ungapped = 1;
				$gapped = 1;
			}
		}
		elsif ($ARGV[$i] =~ /^-+(?:min(?:imum)?ident(?:ity|ities)?|m)=(.+)$/i) {
			my @minident = split(/,/, $1);
			if (scalar(@minident) > 3) {
				&errorMessage(__LINE__, "Too many minident were given.");
			}
			$minident1 = $minident[0];
			if ($minident[1]) {
				$minident2 = $minident[1];
			}
			else {
				$minident2 = $minident1;
			}
			if ($minident[2]) {
				$minident3 = $minident[2];
			}
			else {
				$minident3 = $minident2;
			}
		}
		elsif ($ARGV[$i] =~ /^-+min(?:imum)?(?:overlap|ovl)(?:length|len)=(\d+)$/i) {
			$minovllen = $1;
		}
		elsif ($ARGV[$i] =~ /^-+max(?:imum)?n(?:um)?hits?=(\d+)$/i) {
			$maxnhits = $1;
		}
		elsif ($ARGV[$i] =~ /^-+(?:strand|s)=(forward|plus|single|both|double)$/i) {
			if ($1 =~ /^(?:forward|plus|single)$/i) {
				$strand = ' -s';
			}
		}
		elsif ($ARGV[$i] =~ /^-+merge(?:mode)?=(rough|lazy|draft|rapid|fast|middle|moderate|normal|accurate|slow|strict)$/i) {
			if ($1 =~ /^(?:accurate|slow|strict)$/i) {
				$mergemode = 'accurate';
			}
			elsif ($1 =~ /^(?:middle|moderate|normal)$/i) {
				$mergemode = 'normal';
			}
			elsif ($1 =~ /^(?:draft|rapid|fast)$/i) {
				$mergemode = 'draft';
			}
			elsif ($1 =~ /^(?:rough|lazy)$/i) {
				$mergemode = 'rough';
			}
		}
		elsif ($ARGV[$i] =~ /^-+(?:(?:aln|align|alignment)wiggle|a)=(\d+)$/i) {
			$alnwiggle = $1;
		}
		elsif ($ARGV[$i] =~ /^-+(?:n|n(?:um)?threads?)=(\d+)$/i) {
			$numthreads = $1;
		}
		elsif ($ARGV[$i] =~ /^-+min(?:imum)?n(?:um)?seq(?:uence)?s?=(\d+)$/i) {
			$minnseq = $1;
		}
		elsif ($ARGV[$i] =~ /^-+max(?:imum)?n(?:um)?splits?=(\d+)$/i) {
			$maxnsplit = $1;
		}
		elsif ($ARGV[$i] =~ /^-+noeid$/i) {
			$noeid = 1;
		}
		elsif ($ARGV[$i] =~ /^-+(?:qual|qualfile|q)=(.+)$/i) {
			my @temp = glob($1);
			if (scalar(@temp > 0)) {
				foreach (@temp) {
					$qualfiles{$_} = 1;
				}
			}
			else {
				&errorMessage(__LINE__, "Quality file does not exist.");
			}
		}
		else {
			my @temp = glob($ARGV[$i]);
			if (scalar(@temp > 0)) {
				foreach (@temp) {
					$inputfiles{$_} = 1;
				}
			}
			else {
				&errorMessage(__LINE__, "Input file does not exist.");
			}
		}
	}
	my @temp1 = keys(%inputfiles);
	foreach my $inputfile (@temp1) {
		$inputfile =~ /\.[^\.]+$/;
		my $prefix = $`;
		if (%qualfiles) {
			foreach my $qualfile (keys(%qualfiles)) {
				if ($qualfile =~ /$inputfile/ || $qualfile =~ /^$prefix/) {
					push(@inputfiles, $inputfile);
					push(@qualfiles, $qualfile);
					delete($qualfiles{$qualfile});
					last;
				}
			}
		}
		elsif (-e "$inputfile.qual") {
			push(@inputfiles, $inputfile);
			push(@qualfiles, "$inputfile.qual");
		}
		else {
			push(@inputfiles, $inputfile);
		}
	}
	if (scalar(@inputfiles) != scalar(@temp1) || @qualfiles && scalar(@qualfiles) != scalar(@temp1)) {
		&errorMessage(__LINE__, "Quality file name must contain associated input file name.");
	}
}

# check variables
unless (@inputfiles) {
	&errorMessage(__LINE__, "There is no input files.");
}
if ($minnseq <= 0) {
	&errorMessage(__LINE__, "The minimum number of sequences is too small.");
}
if (defined($maxnsplit) && $maxnsplit < 1) {
	&errorMessage(__LINE__, "The maximum number of splits is too small.");
}
if ($minident1 > 1 || $minident1 < 0.7) {
	&errorMessage(__LINE__, "Minimum identity threshold is invalid.");
}
if ($minident2 > 1 || $minident2 < 0.7) {
	&errorMessage(__LINE__, "Minimum identity threshold is invalid.");
}
if ($minident3 > 1 || $minident3 < 0.7) {
	&errorMessage(__LINE__, "Minimum identity threshold is invalid.");
}
if ($alnwiggle > 100 || $alnwiggle < 2) {
	&errorMessage(__LINE__, "Alignment wiggle value is invalid.");
}
if ($minident1 > $minident2 && ($mergemode eq 'draft' || $mergemode eq 'rough')) {
	&errorMessage(__LINE__, "Fast merging is incompatible to larger minimum identity threshold 1 than minimum identity threshold 2.");
}

# check environment
my $assams = "perl \"$0\"";
my $toAmos;
my $hashoverlap;
my $ungappedoverlap;
my $tigger;
my $makeconsensus;
my $bank2fasta;
my $listReadPlacedStatus;
my $dumpreads;
my $bankreport;
my $banktransact;
my $bankcombine;
{
	my $pathto;
	if ($ENV{'ASSAMSHOME'}) {
		$pathto = $ENV{'ASSAMSHOME'};
	}
	else {
		my $temp;
		if (-e '.assams') {
			$temp = '.assams';
		}
		elsif (-e $ENV{'HOME'} . '/.assams') {
			$temp = $ENV{'HOME'} . '/.assams';
		}
		elsif (-e '/etc/assams/.assams') {
			$temp = '/etc/assams/.assams';
		}
		if ($temp) {
			my $filehandle;
			unless (open($filehandle, "< $temp")) {
				&errorMessage(__LINE__, "Cannot read \"$temp\".");
			}
			while (<$filehandle>) {
				if (/^\s*ASSAMSHOME\s*=\s*(\S[^\r\n]*)/) {
					$pathto = $1;
					$pathto =~ s/\s+$//;
					last;
				}
			}
			close($filehandle);
		}
	}
	if ($pathto) {
		$pathto =~ s/^"(.+)"$/$1/;
		$pathto =~ s/\/$//;
		$pathto .= '/bin';
		if (!-e $pathto) {
			&errorMessage(__LINE__, "Cannot find \"$pathto\".");
		}
		$toAmos = "\"$pathto/toAmos\"";
		$hashoverlap = "\"$pathto/hash-overlap\"";
		$ungappedoverlap = "\"$pathto/ungapped-overlap\"";
		$tigger = "\"$pathto/tigger\"";
		$makeconsensus = "\"$pathto/make-consensus\"";
		$bank2fasta = "\"$pathto/bank2fasta\"";
		$listReadPlacedStatus = "\"$pathto/listReadPlacedStatus\"";
		$dumpreads = "\"$pathto/dumpreads\"";
		$bankreport = "\"$pathto/bank-report\"";
		$banktransact = "\"$pathto/bank-transact\"";
		$bankcombine = "\"$pathto/bank-combine\"";
	}
	else {
		$toAmos = 'toAmos';
		$hashoverlap = 'hash-overlap';
		$ungappedoverlap = 'ungapped-overlap';
		$tigger = 'tigger';
		$makeconsensus = 'make-consensus';
		$bank2fasta = 'bank2fasta';
		$listReadPlacedStatus = 'listReadPlacedStatus';
		$dumpreads = 'dumpreads';
		$bankreport = 'bank-report';
		$banktransact = 'bank-transact';
		$bankcombine = 'bank-combine';
	}
}

# format recognition
my $format;
foreach my $inputfile (@inputfiles) {
	my @temp = glob("$inputfile/*");
	if (@temp && $format eq 'FASTA') {
		&errorMessage(__LINE__, "Input formats are not same.");
	}
	elsif (@temp) {
		$format = 'BANK';
	}
	else {
		$format = 'FASTA';
	}
}

{
	my %nseqs;
	# split sequences
	if ($format eq 'FASTA') {
		print(STDERR "Splitting sequences...");
		my $nseq = 0;
		unless (@qualfiles) {
			print(STDERR "\nWARNING!: line " . __LINE__ . "\nCannot find quality sequence files.\n");
		}
		# check the number of sequences
		my $nqualseq = 0;
		for (my $i = 0; $i < scalar(@inputfiles); $i ++) {
			my $inputhandle;
			unless (open($inputhandle, "< $inputfiles[$i]")) {
				&errorMessage(__LINE__, "Cannot open \"$inputfiles[$i]\".");
			}
			while (<$inputhandle>) {
				if (/^>/) {
					$nseq ++;
				}
			}
			close($inputhandle);
			if (@qualfiles) {
				unless (open($inputhandle, "< $qualfiles[$i]")) {
					&errorMessage(__LINE__, "Cannot open \"$qualfiles[$i]\".");
				}
				while (<$inputhandle>) {
					if (/^>/) {
						$nqualseq ++;
					}
				}
				close($inputhandle);
				if ($nqualseq != $nseq) {
					&errorMessage(__LINE__, "The number of quality sequences is not equal to the number of nucleotide sequences.");
				}
			}
		}
		# calculate nseqs of each file
		my @nseqperfile;
		if ($nseq < $minnseq * 2) {
			push(@nseqperfile, $nseq);
		}
		elsif ($maxnsplit && int($nseq / $minnseq) > $maxnsplit) {
			my $temp = int($nseq / $maxnsplit);
			for (my $i = 0; $i < $maxnsplit; $i ++) {
				if ($i < $nseq % $maxnsplit) {
					push(@nseqperfile, $temp + 1);
				}
				else {
					push(@nseqperfile, $temp);
				}
			}
		}
		else {
			my $nsplit = int($nseq / $minnseq);
			my $temp = int($nseq / $nsplit);
			for (my $i = 0; $i < $nsplit; $i ++) {
				if ($i < $nseq % $nsplit) {
					push(@nseqperfile, $temp + 1);
				}
				else {
					push(@nseqperfile, $temp);
				}
			}
		}
		# make splitted files
		{
			my $tempnseq = 0;
			my $tempfileno = 0;
			my $outputhandle;
			unless (open($outputhandle, "> $outputbank.$tempfileno.fasta")) {
				&errorMessage(__LINE__, "Cannot make \"$outputbank.$tempfileno.fasta\".");
			}
			for (my $i = 0; $i < scalar(@inputfiles); $i ++) {
				my $inputhandle;
				unless (open($inputhandle, "< $inputfiles[$i]")) {
					&errorMessage(__LINE__, "Cannot open \"$inputfiles[$i]\".");
				}
				while (<$inputhandle>) {
					if (/^>/) {
						$tempnseq ++;
						if ($tempnseq > $nseqperfile[$tempfileno]) {
							close($outputhandle);
							$tempfileno ++;
							$tempnseq = 1;
							unless (open($outputhandle, "> $outputbank.$tempfileno.fasta")) {
								&errorMessage(__LINE__, "Cannot make \"$outputbank.$tempfileno.fasta\".");
							}
						}
					}
					print($outputhandle $_);
				}
				close($inputhandle);
			}
			close($outputhandle);
			if (@qualfiles) {
				$tempnseq = 0;
				$tempfileno = 0;
				unless (open($outputhandle, "> $outputbank.$tempfileno.fasta.qual")) {
					&errorMessage(__LINE__, "Cannot make \"$outputbank.$tempfileno.fasta.qual\".");
				}
				for (my $i = 0; $i < scalar(@inputfiles); $i ++) {
					my $inputhandle;
					unless (open($inputhandle, "< $qualfiles[$i]")) {
						&errorMessage(__LINE__, "Cannot open \"$qualfiles[$i]\".");
					}
					while (<$inputhandle>) {
						if (/^>/) {
							$tempnseq ++;
							if ($tempnseq > $nseqperfile[$tempfileno]) {
								close($outputhandle);
								$tempfileno ++;
								$tempnseq = 1;
								unless (open($outputhandle, "> $outputbank.$tempfileno.fasta.qual")) {
									&errorMessage(__LINE__, "Cannot make \"$outputbank.$tempfileno.fasta.qual\".");
								}
							}
						}
						print($outputhandle $_);
					}
					close($inputhandle);
				}
				close($outputhandle);
			}
		}
		# make AMOS bank
		undef(@inputfiles);
		for (my $i = 0; $i < scalar(@nseqperfile); $i ++) {
			if (@qualfiles) {
				if (system("$toAmos -s $outputbank.$i.fasta -q $outputbank.$i.fasta.qual -gq 30 -bq 10 -o - 2> $devnull | $banktransact -c -f -z -b $outputbank.$i.bnk -m - 2> $devnull 1> $devnull")) {
					&errorMessage(__LINE__, "Cannot run \"$toAmos -s $outputbank.$i.fasta -q $outputbank.$i.fasta.qual -gq 30 -bq 10 -o - | $banktransact -c -f -z -b $outputbank.$i.bnk -m -\".");
				}
				unless (unlink("$outputbank.$i.fasta.qual")) {
					&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i.fasta.qual\".");
				}
			}
			else {
				if (system("$toAmos -s $outputbank.$i.fasta -o - 2> $devnull | $banktransact -c -f -z -b $outputbank.$i.bnk -m - 2> $devnull 1> $devnull")) {
					&errorMessage(__LINE__, "Cannot run \"$toAmos -s $outputbank.$i.fasta -o - | $banktransact -c -f -z -b $outputbank.$i.bnk -m -\".");
				}
			}
			unless (unlink("$outputbank.$i.fasta")) {
				&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i.fasta\".");
			}
			push(@inputfiles, "$outputbank.$i.bnk");
			$nseqs{"$outputbank.$i.bnk"} = $nseqperfile[$i];
		}
		print(STDERR "done.\n\n");
	}
	else {
		my $pipehandle;
		for (my $i = 0; $i < scalar(@inputfiles); $i ++) {
			my $nseq = 0;
			unless (open($pipehandle, "$dumpreads -r -e $inputfiles[$i] 2> $devnull |")) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $inputfiles[$i]\".");
			}
			while (<$pipehandle>) {
				if (/^>/) {
					$nseq ++;
				}
			}
			close($pipehandle);
			if ($?) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $inputfiles[$i]\".");
			}
			$nseqs{$inputfiles[$i]} = $nseq;
		}
	}
	# assemble sequences by minimus pipeline in parallel
	print(STDERR "Assembling sequences...\n");
	for (my $i = 0; $i < scalar(@inputfiles); $i ++) {
		if ($nseqs{$inputfiles[$i]} > 1 && !-e "$inputfiles[$i]/CTG.ifo") {
			print(STDERR "Processing $inputfiles[$i]...\n");
			if ($minident1 == 1 || $exact) {
				&assembleBank($ungappedoverlap, 0, 0, $inputfiles[$i]);
			}
			if ($minident1 < 1) {
				my $tempopt;
				if ($strand) {
					$tempopt .= " --strand=plus";
				}
				if ($maxnsplit) {
					$tempopt .= " --maxnsplits=$maxnsplit";
				}
				if ($ungapped) {
					if (-e "$inputfiles[$i]/CTG.ifo") {
						if (system("$assams$tempopt --noeid --assemblemode=ungapped --mergemode=$mergemode --minnseq=$minnseq --minident=1,$minident1,$minident1 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $inputfiles[$i] $inputfiles[$i].temp.bnk 2> $devnull 1> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --assemblemode=ungapped --mergemode=$mergemode --minnseq=$minnseq --minident=1,$minident1,$minident1 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $inputfiles[$i] $inputfiles[$i].temp.bnk\".");
						}
						#if (!-e "$inputfiles[$i].temp.bnk/CTG.ifo") {
						#	&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --assemblemode=ungapped --mergemode=$mergemode --minnseq=$minnseq --minident=1,$minident1,$minident1 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $inputfiles[$i] $inputfiles[$i].temp.bnk\".");
						#}
						&deleteBanks($inputfiles[$i]);
						unless (rename("$inputfiles[$i].temp.bnk", $inputfiles[$i])) {
							&errorMessage(__LINE__, "Cannot rename \"$inputfiles[$i].temp.bnk\" to \"$inputfiles[$i]\".");
						}
					}
					else {
						&assembleBank($ungappedoverlap, (1 - $minident1), ((1 - $minident1) * 2), $inputfiles[$i]);
					}
				}
				if ($gapped) {
					if (-e "$inputfiles[$i]/CTG.ifo") {
						if (system("$assams$tempopt --noeid --assemblemode=gapped --mergemode=$mergemode --minnseq=$minnseq --minident=1,$minident1,$minident1 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $inputfiles[$i] $inputfiles[$i].temp.bnk 2> $devnull 1> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --assemblemode=gapped --mergemode=$mergemode --minnseq=$minnseq --minident=1,$minident1,$minident1 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $inputfiles[$i] $inputfiles[$i].temp.bnk\".");
						}
						#if (!-e "$inputfiles[$i].temp.bnk/CTG.ifo") {
						#	&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --assemblemode=gapped --mergemode=$mergemode --minnseq=$minnseq --minident=1,$minident1,$minident1 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $inputfiles[$i] $inputfiles[$i].temp.bnk\".");
						#}
						&deleteBanks($inputfiles[$i]);
						unless (rename("$inputfiles[$i].temp.bnk", $inputfiles[$i])) {
							&errorMessage(__LINE__, "Cannot rename \"$inputfiles[$i].temp.bnk\" to \"$inputfiles[$i]\".");
						}
					}
					else {
						&assembleBank($hashoverlap, (1 - $minident1), ((1 - $minident1) * 2), $inputfiles[$i]);
					}
				}
			}
		}
	}
	print(STDERR "done.\n\n");
}

if (scalar(@inputfiles) == 1) {
	if ($minident1 == $minident2) {
		print(STDERR "Generating output bank...\n");
		&cleanBank($inputfiles[0], 'CTG', 'RED');
		my $pipehandlein;
		unless (open($pipehandlein, "$bankreport -p -b $inputfiles[0] 2> $devnull |")) {
			&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $inputfiles[0]\".");
		}
		my $pipehandleout;
		unless (open($pipehandleout, "| $banktransact -c -f -z -b $outputbank -m - 1> $devnull 2> $devnull")) {
			&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank -m -\".");
		}
		my $temp;
		my $nesting = 0;
		my $contig;
		while (<$pipehandlein>) {
			if ($nesting && /^\}/) {
				$nesting --;
			}
			elsif (/^\{([A-Z]+)/) {
				$nesting ++;
				if ($1 eq 'CTG') {
					$contig ++;
				}
			}
			$temp .= $_;
			if ($nesting == 0) {
				if ($contig) {
					$temp =~ s/^eid:\S+\r?\n?//m;
					unless ($noeid) {
						$temp =~ s/^iid:(\d+)/$&\neid:contig_$1/m;
					}
				}
				print($pipehandleout $temp);
				undef($temp);
				undef($contig);
			}
		}
		close($pipehandleout);
		if ($?) {
			&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank -m -\".");
		}
		close($pipehandlein);
		if ($?) {
			&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $inputfiles[0]\".");
		}
		#if (!-e "$outputbank/CTG.ifo") {
		#	&errorMessage(__LINE__, "Invalid assemble result.");
		#}
		# delete temporary bank
		if ($format eq 'FASTA') {
			&deleteBanks($inputfiles[0]);
		}
		print(STDERR "done.\n\n");
	}
	else {
		# 2ndary assemble
		print(STDERR "Running secondary assemble...\n");
		if (-e "$inputfiles[0]/CTG.ifo") {
			if (system("$bank2fasta -b $inputfiles[0] -iid -q $outputbank.0.fasta.qual > $outputbank.0.fasta 2> $devnull")) {
				&errorMessage(__LINE__, "Cannot run \"$bank2fasta -b $inputfiles[0] -iid -q $outputbank.0.fasta.qual > $outputbank.0.fasta\".");
			}
			if (system("$listReadPlacedStatus -S -I $inputfiles[0] > $outputbank.0.singletons.iid 2> $devnull")) {
				&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus -S -I $inputfiles[0] > $outputbank.0.singletons.iid\".");
			}
			if (system("$dumpreads -r -e -I $outputbank.0.singletons.iid $inputfiles[0] >> $outputbank.0.fasta 2> $devnull")) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -I $outputbank.0.singletons.iid $inputfiles[0] >> $outputbank.0.fasta\".");
			}
			if (system("$dumpreads -r -e -q -I $outputbank.0.singletons.iid $inputfiles[0] >> $outputbank.0.fasta.qual 2> $devnull")) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -q -I $outputbank.0.singletons.iid $inputfiles[0] >> $outputbank.0.fasta.qual\".");
			}
			unless (unlink("$outputbank.0.singletons.iid")) {
				&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.singletons.iid\".");
			}
		}
		else {
			if (system("$dumpreads -r -e $inputfiles[0] > $outputbank.0.fasta 2> $devnull")) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $inputfiles[0] > $outputbank.0.fasta\".");
			}
			if (system("$dumpreads -r -e -q $inputfiles[0] > $outputbank.0.fasta.qual 2> $devnull")) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -q $inputfiles[0] > $outputbank.0.fasta.qual\".");
			}
		}
		# make combined files
		{
			my $outputhandle1;
			unless (open($outputhandle1, "> $outputbank.temp.fasta")) {
				&errorMessage(__LINE__, "Cannot write \"$outputbank.temp.fasta\".");
			}
			my $outputhandle2;
			unless (open($outputhandle2, "> $outputbank.temp.fasta.qual")) {
				&errorMessage(__LINE__, "Cannot write \"$outputbank.temp.fasta.qual\".");
			}
			my $inputhandle;
			unless (open($inputhandle, "< $outputbank.0.fasta")) {
				&errorMessage(__LINE__, "Cannot read \"$outputbank.0.fasta\".");
			}
			while (<$inputhandle>) {
				s/^>(\d+)(\r?\n?)$/>contig_A\_$1$2/;
				print($outputhandle1 $_);
			}
			close($inputhandle);
			unless (unlink("$outputbank.0.fasta")) {
				&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.fasta\".");
			}
			unless (open($inputhandle, "< $outputbank.0.fasta.qual")) {
				&errorMessage(__LINE__, "Cannot open \"$outputbank.0.fasta.qual\".");
			}
			while (<$inputhandle>) {
				s/^>(\d+)(\r?\n?)$/>contig_A_$1$2/;
				print($outputhandle2 $_);
			}
			close($inputhandle);
			unless (unlink("$outputbank.0.fasta.qual")) {
				&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.fasta.qual\".");
			}
			close($outputhandle1);
			close($outputhandle2);
			unless (rename("$outputbank.temp.fasta", "$outputbank.0.fasta")) {
				&errorMessage(__LINE__, "Cannot rename \"$outputbank.temp.fasta\" to \"$outputbank.0.fasta\".");
			}
			unless (rename("$outputbank.temp.fasta.qual", "$outputbank.0.fasta.qual")) {
				&errorMessage(__LINE__, "Cannot rename \"$outputbank.temp.fasta.qual\" to \"$outputbank.0.fasta.qual\".");
			}
		}
		# secondary assemble
		if (system("$toAmos -s $outputbank.0.fasta -q $outputbank.0.fasta.qual -gq 30 -bq 10 -o - 2> $devnull | $banktransact -c -f -z -b $outputbank.0.temp1.bnk -m - 2> $devnull 1> $devnull")) {
			&errorMessage(__LINE__, "Cannot run \"$toAmos -s $outputbank.0.fasta -q $outputbank.0.fasta.qual -gq 30 -bq 10 -o - | $banktransact -c -f -z -b $outputbank.0.temp1.bnk -m -\".");
		}
		unless (unlink("$outputbank.0.fasta")) {
			&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.fasta\".");
		}
		unless (unlink("$outputbank.0.fasta.qual")) {
			&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.fasta.qual\".");
		}
		if ($minident2 == 1 || $exact) {
			&assembleBank($ungappedoverlap, 0, 0, "$outputbank.0.temp1.bnk");
		}
		elsif ($ungapped) {
			&assembleBank($ungappedoverlap, (1 - $minident2), 0.6, "$outputbank.0.temp1.bnk");
		}
		else {
			&assembleBank($hashoverlap, (1 - $minident2), 0.6, "$outputbank.0.temp1.bnk");
		}
		# read assemble results and get contig-reads association
		my %primaryaffiliation; # affiliation of sequence reads to primary contig
		my %secondaryaffiliation; # affiliation of sequence reads to secondary contig
		my %secondarysingletons;
		my %singletons;
		{
			# read 2ndary assemble results
			my $pipehandle;
			my %temp;
			if (-e "$outputbank.0.temp1.bnk/CTG.ifo") {
				unless (open($pipehandle, "$listReadPlacedStatus $outputbank.0.temp1.bnk 2> $devnull |")) {
					&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $outputbank.0.temp1.bnk\".");
				}
				while (<$pipehandle>) {
					s/\r?\n?$//;
					if (my @row = split(/\t/)) {
						if (scalar(@row) == 4 && $row[2] eq 'S' && $row[1] =~ /^contig_A_(\d+)$/) {
							$secondarysingletons{$1} = $row[0];
						}
						elsif (scalar(@row) == 4 && $row[2] eq 'S') {
							$singletons{$row[1]} = 1;
						}
						elsif (scalar(@row) == 5 && $row[2] eq 'P' && $row[1] =~ /^contig_A_\d+$/) {
							$temp{$&} = $row[4];
						}
						elsif (scalar(@row) == 5 && $row[2] eq 'P') {
							$secondaryaffiliation{$row[1]} = $row[4];
						}
						else {
							&errorMessage(__LINE__, "Invalid assemble results.\nBank: $outputbank.0.temp1.bnk\nSequence: $row[1]\nSecondary contig: $row[4]\n");
						}
					}
				}
				close($pipehandle);
				if ($?) {
					&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $outputbank.0.temp1.bnk\".");
				}
			}
			else {
				unless (open($pipehandle, "$dumpreads -r -e $outputbank.0.temp1.bnk 2> $devnull |")) {
					&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $outputbank.0.temp1.bnk\".");
				}
				while (<$pipehandle>) {
					if (/^>contig_A_(\d+)/) {
						$secondarysingletons{$1} = 1;
					}
					elsif (/^>(\S+)/) {
						$singletons{$1} = 1;
					}
				}
				close($pipehandle);
				if ($?) {
					&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $outputbank.0.temp1.bnk\".");
				}
			}
			# read primary assemble results
			if (-e "$inputfiles[0]/CTG.ifo") {
				unless (open($pipehandle, "$listReadPlacedStatus $inputfiles[0] 2> $devnull |")) {
					&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $inputfiles[0]\".");
				}
				while (<$pipehandle>) {
					s/\r?\n?$//;
					if (my @row = split(/\t/)) {
						if (scalar(@row) == 4 && $row[2] eq 'S') {
							if (!exists($secondaryaffiliation{$row[1]}) && !exists($singletons{$row[1]})) {
								&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[0]\nSequence: $row[1]\n");
							}
						}
						elsif (scalar(@row) == 5 && $row[2] eq 'P' && $temp{"contig_A_$row[4]"}) {
							$secondaryaffiliation{$row[1]} = $temp{"contig_A_$row[4]"};
						}
						elsif (scalar(@row) == 5 && $row[2] eq 'P') {
							$primaryaffiliation{$row[1]} = $row[4];
						}
						else {
							&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[0]\nSequence: $row[1]\nPrimary contig: contig_A_$row[4]\nSecondary contig: " . $temp{"contig_A_$row[4]"} . "\n");
						}
					}
				}
				close($pipehandle);
				if ($?) {
					&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $inputfiles[0]\".");
				}
			}
		}
		# get secondary contigs and all reads, and save them to appropreate files
		my $num2ndcontig = 1;
		{
			# make 2ndcontig files
			# sequence
			# secondarycontig
			my $pipehandle;
			my $outputhandle;
			my $inputhandle;
			if (-e "$outputbank.0.temp1.bnk/CTG.ifo") {
				unless (open($pipehandle, "$bank2fasta -b $outputbank.0.temp1.bnk -iid -q $outputbank.0.2ndcontigs.fasta.qual 2> $devnull |")) {
					&errorMessage(__LINE__, "Cannot run \"$bank2fasta -b $outputbank.0.temp1.bnk -iid -q $outputbank.0.2ndcontigs.fasta.qual\".");
				}
				while (<$pipehandle>) {
					if (s/^>(\d+)(\r?\n?)$/>secondarycontig_$num2ndcontig$2/) {
						if ($1 != $num2ndcontig) {
							&errorMessage(__LINE__, "Invalid assemble results.");
						}
						if ($outputhandle) {
							close($outputhandle);
						}
						unless (open($outputhandle, "> $outputbank.0.2ndcontig$num2ndcontig.fasta")) {
							&errorMessage(__LINE__, "Cannot make \"$outputbank.0.2ndcontig$num2ndcontig.fasta\".");
						}
						$num2ndcontig ++;
					}
					print($outputhandle $_);
				}
				if ($outputhandle) {
					close($outputhandle);
				}
				close($pipehandle);
				if ($?) {
					&errorMessage(__LINE__, "Cannot run \"$bank2fasta -b $outputbank.0.temp1.bnk -iid -q $outputbank.0.2ndcontigs.fasta.qual\".");
				}
				# quality
				$num2ndcontig = 1;
				# secondarycontig
				unless (open($inputhandle, "< $outputbank.0.2ndcontigs.fasta.qual")) {
					&errorMessage(__LINE__, "Cannot open \"$outputbank.0.2ndcontigs.fasta.qual\".");
				}
				while (<$inputhandle>) {
					if (s/^>(\d+)(\r?\n?)$/>secondarycontig_$num2ndcontig$2/) {
						if ($1 != $num2ndcontig) {
							&errorMessage(__LINE__, "Invalid assemble results.");
						}
						if ($outputhandle) {
							close($outputhandle);
						}
						unless (open($outputhandle, "> $outputbank.0.2ndcontig$num2ndcontig.fasta.qual")) {
							&errorMessage(__LINE__, "Cannot make \"$outputbank.0.2ndcontig$num2ndcontig.fasta.qual\".");
						}
						$num2ndcontig ++;
					}
					print($outputhandle $_);
				}
				if ($outputhandle) {
					close($outputhandle);
				}
				close($inputhandle);
				unless (unlink("$outputbank.0.2ndcontigs.fasta.qual")) {
					&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.2ndcontigs.fasta.qual\".");
				}
			}
			# make reads files
			unless (open($pipehandle, "$dumpreads -r -e $inputfiles[0] 2> $devnull |")) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $inputfiles[0]\".");
			}
			while (<$pipehandle>) {
				if (/^>(.+)\r?\n?/) {
					my $seqname = $1;
					if ($secondaryaffiliation{$seqname}) {
						if ($outputhandle) {
							close($outputhandle);
						}
						if ($secondaryaffiliation{$seqname} =~ /^\d+$/) {
							unless (open($outputhandle, ">> $outputbank.0.reads$secondaryaffiliation{$seqname}.fasta")) {
								&errorMessage(__LINE__, "Cannot write \"$outputbank.reads$secondaryaffiliation{$seqname}.fasta\".");
							}
						}
						else {
							&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[0]\nSequence: $seqname\nSecondary contig: $secondaryaffiliation{$seqname}\n");
						}
					}
					elsif ($primaryaffiliation{$seqname}) {
						if ($outputhandle) {
							close($outputhandle);
						}
						if ($primaryaffiliation{$seqname} =~ /^\d+$/) {
							unless (open($outputhandle, "> $devnull")) {
								&errorMessage(__LINE__, "Cannot write \"$devnull\".");
							}
						}
						else {
							&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[0]\nSequence: $seqname\nPrimary contig: $primaryaffiliation{$seqname}\n");
						}
					}
					elsif ($singletons{$seqname}) {
						if ($outputhandle) {
							close($outputhandle);
						}
						unless (open($outputhandle, ">> $outputbank.0.singletons.fasta")) {
							&errorMessage(__LINE__, "Cannot write \"$outputbank.0.singletons.fasta\".");
						}
					}
					else {
						&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[0]\nSequence: $seqname\nUnplaced\n");
					}
				}
				print($outputhandle $_);
			}
			if ($outputhandle) {
				close($outputhandle);
			}
			close($pipehandle);
			if ($?) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $inputfiles[0]\".");
			}
			# make quality file
			unless (open($pipehandle, "$dumpreads -r -e -q $inputfiles[0] 2> $devnull |")) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -q $inputfiles[0]\".");
			}
			while (<$pipehandle>) {
				if (/^>(.+)\r?\n?/) {
					my $seqname = $1;
					if ($secondaryaffiliation{$seqname}) {
						if ($outputhandle) {
							close($outputhandle);
						}
						if ($secondaryaffiliation{$seqname} =~ /^\d+$/) {
							unless (open($outputhandle, ">> $outputbank.0.reads$secondaryaffiliation{$seqname}.fasta.qual")) {
								&errorMessage(__LINE__, "Cannot write \"$outputbank.0.reads$secondaryaffiliation{$seqname}.fasta.qual\".");
							}
						}
						else {
							&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[0]\nSequence: $seqname\nSecondary contig: $secondaryaffiliation{$seqname}\n");
						}
					}
					elsif ($primaryaffiliation{$seqname}) {
						if ($outputhandle) {
							close($outputhandle);
						}
						if ($primaryaffiliation{$seqname} =~ /^\d+$/) {
							unless (open($outputhandle, "> $devnull")) {
								&errorMessage(__LINE__, "Cannot write \"$devnull\".");
							}
						}
						else {
							&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[0]\nSequence: $seqname\nPrimary contig: $primaryaffiliation{$seqname}\n");
						}
					}
					elsif ($singletons{$seqname}) {
						if ($outputhandle) {
							close($outputhandle);
						}
						unless (open($outputhandle, ">> $outputbank.0.singletons.fasta.qual")) {
							&errorMessage(__LINE__, "Cannot write \"$outputbank.0.singletons.fasta.qual\".");
						}
					}
					else {
						&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[0]\nSequence: $seqname\nUnplaced\n");
					}
				}
				print($outputhandle $_);
			}
			if ($outputhandle) {
				close($outputhandle);
			}
			close($pipehandle);
			if ($?) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -q $inputfiles[0]\".");
			}
			# make secondarysingleton/primarycontig bank
			if (%secondarysingletons && %primaryaffiliation) {
				my $pipehandlein;
				unless (open($pipehandlein, "$bankreport -p -b $inputfiles[0] 2> $devnull |")) {
					&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $inputfiles[0]\".");
				}
				my $pipehandleout;
				unless (open($pipehandleout, "| $banktransact -c -f -z -b $outputbank.0.reads$num2ndcontig.bnk -m - 2> $devnull 1> $devnull")) {
					&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank.0.reads$num2ndcontig.bnk -m -\".");
				}
				my $temp;
				my $nesting = 0;
				my $contig;
				my $read;
				my $numcontig = 1;
				while (<$pipehandlein>) {
					if ($nesting && /^\}/) {
						$nesting --;
					}
					elsif (/^\{([A-Z]+)/) {
						$nesting ++;
						if ($1 eq 'CTG') {
							$contig ++;
						}
						elsif ($1 eq 'RED') {
							$read ++;
						}
					}
					if ($contig || $read) {
						$temp .= $_;
					}
					if ($temp && $nesting == 0) {
						if ($contig && $temp =~ s/^iid:(\d+)/iid:$numcontig/m && $secondarysingletons{$1}) {
							$temp =~ s/^eid:\S+\r?\n?//m;
							print($pipehandleout $temp);
							$numcontig ++;
						}
						elsif ($read && $temp =~ /^eid:(\S+)/m && $primaryaffiliation{$1}) {
							print($pipehandleout $temp);
						}
						undef($temp);
						undef($contig);
						undef($read);
					}
				}
				close($pipehandleout);
				if ($?) {
					&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank.0.reads$num2ndcontig.bnk -m -\".");
				}
				close($pipehandlein);
				if ($?) {
					&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $inputfiles[0]\".");
				}
			}
		}
		# mapping reads to secondary contig
		{
			my $child = 0;
			$| = 1;
			$? = 0;
			for (my $j = 1; $j < $num2ndcontig; $j ++) {
				if (my $pid = fork()) {
					$child ++;
					if ($child >= ($numthreads + 1)) {
						if (wait == -1) {
							$child = 0;
						} else {
							$child --;
						}
					}
					if ($?) {
						&errorMessage(__LINE__);
					}
					next;
				}
				else {
					if (!-e "$outputbank.0.2ndcontig$j.fasta") {
						&errorMessage(__LINE__, "Cannot find \"$outputbank.0.2ndcontig$j.fasta\".")
					}
					if (!-e "$outputbank.0.2ndcontig$j.fasta.qual") {
						&errorMessage(__LINE__, "Cannot find \"$outputbank.0.2ndcontig$j.fasta.qual\".");
					}
					if (system("$toAmos -s $outputbank.0.2ndcontig$j.fasta -q $outputbank.0.2ndcontig$j.fasta.qual -gq 30 -bq 10 -o - 2> $devnull | $banktransact -c -f -z -b $outputbank.0.reads$j.bnk -m - 2> $devnull 1> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$toAmos -s $outputbank.0.2ndcontig$j.fasta -q $outputbank.0.2ndcontig$j.fasta.qual -gq 30 -bq 10 -o - | $banktransact -c -f -z -b $outputbank.0.reads$j.bnk -m -\".");
					}
					unless (unlink("$outputbank.0.2ndcontig$j.fasta")) {
						&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.2ndcontig$j.fasta\"");
					}
					unless (unlink("$outputbank.0.2ndcontig$j.fasta.qual")) {
						&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.2ndcontig$j.fasta.qual\"");
					}
					&cleanBank("$outputbank.0.reads$j.bnk", 'RED');
					if (!-e "$outputbank.0.reads$j.fasta") {
						&errorMessage(__LINE__, "Cannot find \"$outputbank.0.reads$j.fasta\".")
					}
					if (!-e "$outputbank.0.reads$j.fasta.qual") {
						&errorMessage(__LINE__, "Cannot find \"$outputbank.0.reads$j.fasta.qual\".");
					}
					if (system("$toAmos -s $outputbank.0.reads$j.fasta -q $outputbank.0.reads$j.fasta.qual -gq 30 -bq 10 -id 2 -o - 2> $devnull | $banktransact -z -b $outputbank.0.reads$j.bnk -m - 2> $devnull 1> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$toAmos -s $outputbank.0.reads$j.fasta -q $outputbank.0.reads$j.fasta.qual -gq 30 -bq 10 -id 2 -o - | $banktransact -z -b $outputbank.0.reads$j.bnk -m -\".");
					}
					if (!$ungapped && !$gapped) {
						if (system("OMP_NUM_THREADS=1 $ungappedoverlap$strand -B -r 1 -o $minovllen -m 0 -x 0 -v 0 $outputbank.0.reads$j.bnk 2> $devnull 1> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"OMP_NUM_THREADS=1 $ungappedoverlap$strand -B -r 1 -o $minovllen -m 0 -x 0 -v 0 $outputbank.0.reads$j.bnk\".");
						}
					}
					elsif (!$gapped) {
						if (system("OMP_NUM_THREADS=1 $ungappedoverlap$strand -B -r 1 -o $minovllen -m 0 -x 0.5 -v 0 $outputbank.0.reads$j.bnk 2> $devnull 1> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"OMP_NUM_THREADS=1 $ungappedoverlap$strand -B -r 1 -o $minovllen -m 0 -x 0.5 -v 0 $outputbank.0.reads$j.bnk\".");
						}
					}
					else {
						if (system("OMP_NUM_THREADS=1 $hashoverlap$strand -B -r 1 -o $minovllen -m 0 -x 0.5 -v 0 $outputbank.0.reads$j.bnk 2> $devnull 1> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"OMP_NUM_THREADS=1 $hashoverlap$strand -B -r 1 -o $minovllen -m 0 -x 0.5 -v 0 $outputbank.0.reads$j.bnk\".");
						}
					}
					if (system("$tigger -v 0 -b $outputbank.0.reads$j.bnk 2> $devnull 1> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$tigger -v 0 -b $outputbank.0.reads$j.bnk\".");
					}
					&cleanBank("$outputbank.0.reads$j.bnk", 'LAY', 'RED');
					if (system("$bankreport -p -b $outputbank.0.reads$j.bnk > $outputbank.0.reads$j.afg 2> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $outputbank.0.reads$j.bnk > $outputbank.0.reads$j.afg\".");
					}
					&deleteBanks("$outputbank.0.reads$j.bnk");
					{
						my $inputhandle;
						unless (open($inputhandle, "< $outputbank.0.reads$j.afg")) {
							&errorMessage(__LINE__, "Cannot read \"$outputbank.0.reads$j.afg\".");
						}
						my $pipehandleout;
						unless (open($pipehandleout, "| $banktransact -c -f -z -b $outputbank.0.reads$j.bnk -m - 1> $devnull 2> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank.0.reads$j.bnk -m -\".");
						}
						my $temp;
						my $nesting = 0;
						my $tle;
						my $read;
						while (<$inputhandle>) {
							if ($nesting && /^\}/) {
								$nesting --;
							}
							elsif (/^\{(TLE|RED)/) {
								$nesting ++;
								if ($1 eq 'TLE') {
									$tle ++;
								}
								elsif ($1 eq 'RED') {
									$read ++;
								}
							}
							elsif ($nesting == 0) {
								print($pipehandleout $_);
							}
							if ($tle || $read) {
								$temp .= $_;
							}
							if ($temp && $nesting == 0) {
								if ($tle && $temp !~ /^src:1\r?\n?$/m) {
									print($pipehandleout $temp);
								}
								elsif ($read && $temp !~ /^iid:1\r?\n?$/m) {
									print($pipehandleout $temp);
								}
								undef($temp);
								undef($tle);
								undef($read);
							}
						}
						close($pipehandleout);
						if ($?) {
							&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank.0.reads$j.bnk -m -\".");
						}
						close($inputhandle);
						unless (unlink("$outputbank.0.reads$j.afg")) {
							&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.reads$j.afg\"");
						}
					}
					# if mapping failed
					if (system("$makeconsensus -b -B -o $minovllen -e " . ((1 - $minident3) * 2) . " -w $alnwiggle -v 0 -x $outputbank.0.reads$j.txt $outputbank.0.reads$j.bnk 2> $devnull 1> $devnull")) {
						print(STDERR "Mapping failed at \"$outputbank.0.reads$j.bnk\". Assembly will be retried.\n");
						&deleteBanks("$outputbank.0.reads$j.bnk");
					}
					# if mapping succeeded
					else {
						my %expels;
						my $inputhandle;
						unless (open($inputhandle, "< $outputbank.0.reads$j.txt")) {
							&errorMessage(__LINE__, "Cannot read \"$outputbank.0.reads$j.txt\".");
						}
						while (<$inputhandle>) {
							if (/^(\d+)\s+\d+/) {
								$expels{$1} = 1;
							}
						}
						close($inputhandle);
						# if there are expelled sequences, make new bank
						if (%expels) {
							my $outputhandle;
							unless (open($outputhandle, "> $outputbank.0.reads$j.expels.iid")) {
								&errorMessage(__LINE__, "Cannot make \"$outputbank.0.reads$j.expels.iid\".");
							}
							foreach my $iid (keys(%expels)) {
								print($outputhandle "$iid\n");
							}
							close($outputhandle);
							if (system("$dumpreads -r -e -I $outputbank.0.reads$j.expels.iid $outputbank.0.reads$j.bnk > $outputbank.0.reads$j.expels.fasta 2> $devnull")) {
								&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -I $outputbank.0.reads$j.expels.iid $outputbank.0.reads$j.bnk > $outputbank.0.reads$j.expels.fasta\".");
							}
							if (system("$dumpreads -r -e -q -I $outputbank.0.reads$j.expels.iid $outputbank.0.reads$j.bnk > $outputbank.0.reads$j.expels.fasta.qual 2> $devnull")) {
								&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -q -I $outputbank.0.reads$j.expels.iid $outputbank.0.reads$j.bnk > $outputbank.0.reads$j.expels.fasta.qual\".");
							}
							unless (unlink("$outputbank.0.reads$j.expels.iid")) {
								&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.reads$j.expels.iid\".");
							}
							my $pipehandlein;
							unless (open($pipehandlein, "$bankreport -p -b $outputbank.0.reads$j.bnk 2> $devnull |")) {
								&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $outputbank.0.reads$j.bnk\".");
							}
							my $pipehandleout;
							unless (open($pipehandleout, "| $banktransact -c -f -z -b $outputbank.0.reads$j.temp.bnk -m - 1> $devnull 2> $devnull")) {
								&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank.0.reads$j.temp.bnk -m -\".");
							}
							my $temp;
							my $nesting = 0;
							my $contig;
							my $read;
							while (<$pipehandlein>) {
								if ($nesting && /^\}/) {
									$nesting --;
								}
								elsif (/^\{([A-Z]+)/) {
									$nesting ++;
									if ($1 eq 'CTG') {
										$contig ++;
									}
									elsif ($1 eq 'RED') {
										$read ++;
									}
								}
								#elsif ($nesting == 0) {
								#	print($pipehandleout $_);
								#}
								if ($contig || $read) {
									$temp .= $_;
								}
								if ($temp && $nesting == 0) {
									if ($contig || ($read && $temp =~ /^iid:(\d+)\r?\n?$/m && !$expels{$1})) {
										print($pipehandleout $temp);
									}
									undef($temp);
									undef($contig);
									undef($read);
								}
							}
							close($pipehandleout);
							if ($?) {
								&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank.0.reads$j.temp.bnk -m -\".");
							}
							close($pipehandlein);
							if ($?) {
								&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $outputbank.0.reads$j.bnk\".");
							}
							&deleteBanks("$outputbank.0.reads$j.bnk");
							unless (rename("$outputbank.0.reads$j.temp.bnk", "$outputbank.0.reads$j.bnk")) {
								&errorMessage(__LINE__, "Cannot rename \"$outputbank.0.reads$j.temp.bnk\" to \"$outputbank.0.reads$j.bnk\".");
							}
						}
					}
					exit;
				}
			}
		}
		# join
		while (wait != -1) {
			if ($?) {
				&errorMessage(__LINE__, 'Cannot run merging pipeline correctly.');
			}
		}
		my $tempopt;
		if ($strand) {
			$tempopt .= " --strand=plus";
		}
		if ($maxnsplit) {
			$tempopt .= " --maxnsplits=$maxnsplit";
		}
		if ($exact && $ungapped && $gapped) {
			$tempopt .= " --assemblemode=all";
		}
		elsif ($exact && $ungapped) {
			$tempopt .= " --assemblemode=e+u";
		}
		elsif ($exact && $gapped) {
			$tempopt .= " --assemblemode=e+g";
		}
		elsif ($ungapped && $gapped) {
			$tempopt .= " --assemblemode=u+g";
		}
		elsif ($exact) {
			$tempopt .= " --assemblemode=exact";
		}
		elsif ($ungapped) {
			$tempopt .= " --assemblemode=ungapped";
		}
		elsif ($gapped) {
			$tempopt .= " --assemblemode=gapped";
		}
		if ($minident3 == 1) {
			$tempopt .= " --mergemode=rough";
		}
		else {
			$tempopt .= " --mergemode=$mergemode";
		}
		for (my $j = 1; $j < $num2ndcontig; $j ++) {
			# retry assemble if mapping failed
			if (!-e "$outputbank.0.reads$j.bnk" && -e "$outputbank.0.reads$j.fasta" && -e "$outputbank.0.reads$j.fasta.qual") {
				print(STDERR "Retrying assembly of \"$outputbank.0.reads$j.bnk\"...");
				if (system("$assams$tempopt --noeid --minnseq=$minnseq --minident=$minident3 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.0.reads$j.fasta $outputbank.0.reads$j.bnk 2> $devnull 1> $devnull")) {
					&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --minnseq=$minnseq --minident=$minident3 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.0.reads$j.fasta $outputbank.0.reads$j.bnk\".");
				}
				#if (!-e "$outputbank.0.reads$j.bnk/CTG.ifo") {
				#	&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --minnseq=$minnseq --minident=$minident3 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.0.reads$j.fasta $outputbank.0.reads$j.bnk\".");
				#}
				#if (!-e "$outputbank.0.reads$j.bnk/CTG.ifo" && !-e "$outputbank.0.reads$j.bnk/LAY.ifo" && !-e "$outputbank.0.reads$j.bnk/OVL.ifo") {
				#	if (system("OMP_NUM_THREADS=$numthreads $hashoverlap$strand -B -o $minovllen -m $maxnhits -x " . (1 - $minident3) . " -v 0 $outputbank.0.reads$j.bnk 2> $devnull 1> $devnull")) {
				#		&errorMessage(__LINE__, "Cannot run \"OMP_NUM_THREADS=$numthreads $hashoverlap$strand -B -o $minovllen -m $maxnhits -x " . (1 - $minident3) . " -v 0 $outputbank.0.reads$j.bnk 2> $devnull 1> $devnull\".");
				#	}
				#}
				#if (!-e "$outputbank.0.reads$j.bnk/CTG.ifo" && !-e "$outputbank.0.reads$j.bnk/LAY.ifo") {
				#	if (system("$tigger -v 0 -b $outputbank.0.reads$j.bnk 2> $devnull 1> $devnull")) {
				#		&errorMessage(__LINE__, "Cannot run \"$tigger -v 0 -b $outputbank.0.reads$j.bnk\".");
				#	}
				#}
				#if (!-e "$outputbank.0.reads$j.bnk/CTG.ifo") {
				#	if (system("$makeconsensus -b -B -o $minovllen -e " . ((1 - $minident3) * 2) . " -w $alnwiggle -v 0 $outputbank.0.reads$j.bnk 2> $devnull 1> $devnull")) {
				#		&errorMessage(__LINE__, "Cannot run \"$makeconsensus -b -B -o $minovllen -e " . ((1 - $minident3) * 2) . " -w $alnwiggle -v 0 $outputbank.0.reads$j.bnk\".");
				#	}
				#}
				print(STDERR "OK.\n");
			}
			elsif (!-e "$outputbank.0.reads$j.bnk" || !-e "$outputbank.0.reads$j.bnk/CTG.ifo" || !-e "$outputbank.0.reads$j.bnk/RED.ifo") {
				&errorMessage(__LINE__, "\"$outputbank.0.reads$j.bnk\" is invalid.");
			}
			if (-e "$outputbank.0.reads$j.txt") {
				unless (unlink("$outputbank.0.reads$j.txt")) {
					&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.reads$j.txt\"");
				}
			}
			if (-e "$outputbank.0.reads$j.fasta") {
				unless (unlink("$outputbank.0.reads$j.fasta")) {
					&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.reads$j.fasta\"");
				}
			}
			if (-e "$outputbank.0.reads$j.fasta.qual") {
				unless (unlink("$outputbank.0.reads$j.fasta.qual")) {
					&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.reads$j.fasta.qual\"");
				}
			}
			# assemble expelled sequences
			if (-e "$outputbank.0.reads$j.expels.fasta" && -e "$outputbank.0.reads$j.expels.fasta.qual") {
				print(STDERR "Assembling expelled sequences from \"$outputbank.0.reads$j.bnk\"...");
				if (system("$assams$tempopt --noeid --minnseq=$minnseq --minident=$minident3 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.0.reads$j.expels.fasta $outputbank.0.reads$j.expels.bnk 2> $devnull 1> $devnull")) {
					&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --minnseq=$minnseq --minident=$minident3 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.0.reads$j.expels.fasta $outputbank.0.reads$j.expels.bnk\".");
				}
				#if (!-e "$outputbank.0.reads$j.expels.bnk/CTG.ifo") {
				#	&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --minnseq=$minnseq --minident=$minident3 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.0.reads$j.expels.fasta $outputbank.0.reads$j.expels.bnk\".");
				#}
				#if (!-e "$outputbank.0.reads$j.expels.bnk/CTG.ifo" && !-e "$outputbank.0.reads$j.expels.bnk/LAY.ifo" && !-e "$outputbank.0.reads$j.expels.bnk/OVL.ifo") {
				#	if (system("OMP_NUM_THREADS=$numthreads $hashoverlap$strand -B -o $minovllen -m $maxnhits -x " . (1 - $minident3) . " -v 0 $outputbank.0.reads$j.expels.bnk 2> $devnull 1> $devnull")) {
				#		&errorMessage(__LINE__, "Cannot run \"OMP_NUM_THREADS=$numthreads $hashoverlap$strand -B -o $minovllen -m $maxnhits -x " . (1 - $minident3) . " -v 0 $outputbank.0.reads$j.expels.bnk 2> $devnull 1> $devnull\".");
				#	}
				#}
				#if (!-e "$outputbank.0.reads$j.expels.bnk/CTG.ifo" && !-e "$outputbank.0.reads$j.expels.bnk/LAY.ifo") {
				#	if (system("$tigger -v 0 -b $outputbank.0.reads$j.expels.bnk 2> $devnull 1> $devnull")) {
				#		&errorMessage(__LINE__, "Cannot run \"$tigger -v 0 -b $outputbank.0.reads$j.expels.bnk\".");
				#	}
				#}
				#if (!-e "$outputbank.0.reads$j.expels.bnk/CTG.ifo") {
				#	if (system("$makeconsensus -b -B -o $minovllen -e " . ((1 - $minident3) * 2) . " -w $alnwiggle -v 0 $outputbank.0.reads$j.expels.bnk 2> $devnull 1> $devnull")) {
				#		&errorMessage(__LINE__, "Cannot run \"$makeconsensus -b -B -o $minovllen -e " . ((1 - $minident3) * 2) . " -w $alnwiggle -v 0 $outputbank.0.reads$j.expels.bnk\".");
				#	}
				#}
				unless (unlink("$outputbank.0.reads$j.expels.fasta")) {
					&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.reads$j.expels.fasta\"");
				}
				unless (unlink("$outputbank.0.reads$j.expels.fasta.qual")) {
					&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.reads$j.expels.fasta.qual\"");
				}
				print(STDERR "OK.\n");
			}
		}
		# make combined bank
		{
			if (-e "$outputbank.0.reads1.bnk") {
				my @readsbanks = glob("$outputbank.0.reads*.bnk");
				if (scalar(@readsbanks) > 1) {
					my ($placed, $unplaced) = &sortBanks(@readsbanks);
					@readsbanks = @{$placed};
					foreach my $readsbank (@{$unplaced}) {
						if (system("$dumpreads -r -e $readsbank >> $outputbank.0.singletons.fasta 2> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $readsbank >> $outputbank.0.singletons.fasta\".");
						}
						if (system("$dumpreads -r -e -q $readsbank >> $outputbank.0.singletons.fasta.qual 2> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -q $readsbank >> $outputbank.0.singletons.fasta.qual\".");
						}
						&deleteBanks($readsbank);
					}
					my $tempbank = shift(@readsbanks);
					while (@readsbanks) {
						my @tempbanks = splice(@readsbanks, 0, 20);
						if (system("$bankcombine $outputbank.0.temp2.bnk $tempbank @tempbanks")) {
							&errorMessage(__LINE__, "Cannot run \"$bankcombine $outputbank.0.temp2.bnk $tempbank @tempbanks\".");
						}
						&deleteBanks($tempbank, @tempbanks);
						if (@readsbanks) {
							unless (rename("$outputbank.0.temp2.bnk", $tempbank)) {
								&errorMessage(__LINE__, "Cannot rename \"$outputbank.0.temp2.bnk\" to \"$tempbank\".");
							}
						}
					}
				}
				else {
					unless (rename("$outputbank.0.reads1.bnk", "$outputbank.0.temp2.bnk")) {
						&errorMessage(__LINE__, "Cannot rename \"$outputbank.0.reads1.bnk\" to \"$outputbank.0.temp2.bnk\".");
					}
				}
			}
			if (-e "$outputbank.0.temp2.bnk") {
				my $maxiid = 1;
				my $pipehandlein;
				unless (open($pipehandlein, "$bankreport -p -b $outputbank.0.temp2.bnk 2> $devnull |")) {
					&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $outputbank.0.temp2.bnk\".");
				}
				my $pipehandleout;
				unless (open($pipehandleout, "| $banktransact -c -f -z -b $outputbank -m - 1> $devnull 2> $devnull")) {
					&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank -m -\".");
				}
				my $temp;
				my $nesting = 0;
				my $contig;
				my $read;
				while (<$pipehandlein>) {
					if (/^iid:(\d+)/ && $1 > $maxiid) {
						$maxiid = $1;
					}
					if ($nesting && /^\}/) {
						$nesting --;
					}
					elsif (/^\{([A-Z]+)/) {
						$nesting ++;
						if ($1 eq 'CTG') {
							$contig ++;
						}
						elsif ($1 eq 'RED') {
							$read ++;
						}
					}
					if ($contig || $read) {
						$temp .= $_;
					}
					if ($temp && $nesting == 0) {
						if ($contig) {
							$temp =~ s/^eid:\S+\r?\n?//m;
							unless ($noeid) {
								$temp =~ s/^iid:(\d+)/$&\neid:contig_$1/m;
							}
						}
						print($pipehandleout $temp);
						undef($temp);
						undef($contig);
						undef($read);
					}
				}
				close($pipehandleout);
				if ($?) {
					&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank -m -\".");
				}
				close($pipehandlein);
				if ($?) {
					&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $outputbank.0.temp2.bnk\".");
				}
				# delete temporary bank
				&deleteBanks("$outputbank.0.temp2.bnk");
				&cleanBank($outputbank, 'CTG', 'RED');
				$maxiid ++;
				if (-e "$outputbank.0.singletons.fasta" && -e "$outputbank.0.singletons.fasta.qual") {
					if (system("$toAmos -s $outputbank.0.singletons.fasta -q $outputbank.0.singletons.fasta.qual -gq 30 -bq 10 -id $maxiid -o - 2> $devnull | $banktransact -z -b $outputbank -m - 2> $devnull 1> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$toAmos -s $outputbank.0.singletons.fasta -q $outputbank.0.singletons.fasta.qual -gq 30 -bq 10 -id $maxiid -o - | $banktransact -z -b $outputbank -m -\".");
					}
					unless (unlink("$outputbank.0.singletons.fasta")) {
						&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.singletons.fasta\".");
					}
					unless (unlink("$outputbank.0.singletons.fasta.qual")) {
						&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.singletons.fasta.qual\".");
					}
				}
			}
			elsif (-e "$outputbank.0.singletons.fasta" && -e "$outputbank.0.singletons.fasta.qual") {
				if (system("$toAmos -s $outputbank.0.singletons.fasta -q $outputbank.0.singletons.fasta.qual -gq 30 -bq 10 -o - 2> $devnull | $banktransact -c -f -z -b $outputbank -m - 2> $devnull 1> $devnull")) {
					&errorMessage(__LINE__, "Cannot run \"$toAmos -s $outputbank.0.singletons.fasta -q $outputbank.0.singletons.fasta.qual -gq 30 -bq 10 -o - | $banktransact -c -f -z -b $outputbank -m -\".");
				}
				unless (unlink("$outputbank.0.singletons.fasta")) {
					&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.singletons.fasta\".");
				}
				unless (unlink("$outputbank.0.singletons.fasta.qual")) {
					&errorMessage(__LINE__, "Cannot delete \"$outputbank.0.singletons.fasta.qual\".");
				}
			}
			else {
				unless (dircopy($inputfiles[0], $outputbank)) {
					&errorMessage(__LINE__, "Cannot copy \"$inputfiles[0]\" to \"$outputbank\".");
				}
			}
		}
		# delete temporary bank
		if ($inputfiles[0] =~ /^$outputbank\.\d+\.bnk$/) {
			&deleteBanks($inputfiles[0]);
		}
		# delete temporary bank
		&deleteBanks("$outputbank.0.temp1.bnk");
		&cleanBank($outputbank, 'CTG', 'RED');
		print(STDERR "done.\n\n");
	}
}
else {
	# progressive merge process
	print(STDERR "Merging assemblies progressively...\n");
	my $step = 1;
	while (scalar(@inputfiles) > 1) {
		print(STDERR "Step $step...\n");
		# determine merge combination
		if (scalar(@inputfiles) > 2) {
			# count the number of contigs+singletons
			my %ncontigs;
			for (my $i = 0; $i < scalar(@inputfiles); $i ++) {
				if (-e "$inputfiles[$i]/CTG.ifo") {
					my $ncontigs = 0;
					my %temp;
					my $pipehandle;
					unless (open($pipehandle, "$listReadPlacedStatus $inputfiles[$i] 2> $devnull |")) {
						&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $inputfiles[$i]\".");
					}
					while (<$pipehandle>) {
						if (my @row = split(/\t/)) {
							if (scalar(@row) == 4 && $row[2] eq 'S') {
								$ncontigs ++;
							}
							elsif (scalar(@row) == 5 && $row[2] eq 'P' && !$temp{"$row[4]"}) {
								$temp{"$row[4]"} = 1;
								$ncontigs ++;
							}
						}
					}
					close($pipehandle);
					if ($?) {
						&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $inputfiles[$i]\".");
					}
					if ($ncontigs) {
						$ncontigs{$inputfiles[$i]} = $ncontigs;
					}
					else {
						&errorMessage(__LINE__, "Cannot find contigs and singletons at \"$inputfiles[$i]\".");
					}
				}
				else {
					my $nseq = 0;
					my $pipehandle;
					unless (open($pipehandle, "$dumpreads -r -e $inputfiles[$i] 2> $devnull |")) {
						&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $inputfiles[$i]\".");
					}
					while (<$pipehandle>) {
						if (/^>/) {
							$nseq ++;
						}
					}
					close($pipehandle);
					if ($?) {
						&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $inputfiles[$i]\".");
					}
					if ($nseq) {
						$ncontigs{$inputfiles[$i]} = $nseq;
					}
					else {
						&errorMessage(__LINE__, "Cannot find contigs and singletons at \"$inputfiles[$i]\".");
					}
				}
			}
			my @temp = sort({$ncontigs{$a} <=> $ncontigs{$b}} keys(%ncontigs));
			undef(@inputfiles);
			# set Nth large and Nth small banks to the pair
			while (@temp) {
				if (scalar(@temp == 1)) {
					push(@inputfiles, shift(@temp));
				}
				else {
					push(@inputfiles, shift(@temp), pop(@temp));
				}
			}
		}
		# merge assemblies in parallel
		for (my $i = 0; $i < scalar(@inputfiles) - 1; $i += 2) {
			print(STDERR "Merging $i+" . ($i + 1) . "...\n");
			# output contigs and singletons
			for (my $j = $i; $j <= $i + 1; $j ++) {
				if (-e "$inputfiles[$j]/CTG.ifo") {
					if (system("$bank2fasta -b $inputfiles[$j] -iid -q $outputbank.$j.fasta.qual > $outputbank.$j.fasta 2> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$bank2fasta -b $inputfiles[$j] -iid -q $outputbank.$j.fasta.qual > $outputbank.$j.fasta\".");
					}
					if (system("$listReadPlacedStatus -S -I $inputfiles[$j] > $outputbank.$j.singletons.iid 2> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus -S -I $inputfiles[$j] > $outputbank.$j.singletons.iid\".");
					}
					if (system("$dumpreads -r -e -I $outputbank.$j.singletons.iid $inputfiles[$j] >> $outputbank.$j.fasta 2> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -I $outputbank.$j.singletons.iid $inputfiles[$j] >> $outputbank.$j.fasta\".");
					}
					if (system("$dumpreads -r -e -q -I $outputbank.$j.singletons.iid $inputfiles[$j] >> $outputbank.$j.fasta.qual 2> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -q -I $outputbank.$j.singletons.iid $inputfiles[$j] >> $outputbank.$j.fasta.qual\".");
					}
					unless (unlink("$outputbank.$j.singletons.iid")) {
						&errorMessage(__LINE__, "Cannot delete \"$outputbank.$j.singletons.iid\".");
					}
				}
				else {
					if (system("$dumpreads -r -e $inputfiles[$j] > $outputbank.$j.fasta 2> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $inputfiles[$j] > $outputbank.$j.fasta\".");
					}
					if (system("$dumpreads -r -e -q $inputfiles[$j] > $outputbank.$j.fasta.qual 2> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -q $inputfiles[$j] > $outputbank.$j.fasta.qual\".");
					}
				}
			}
			# make combined files
			my $nseqfirst = 0;
			{
				my $outputhandle1;
				unless (open($outputhandle1, "> $outputbank.$i." . ($i + 1) . ".fasta")) {
					&errorMessage(__LINE__, "Cannot write \"$outputbank.$i." . ($i + 1) . ".fasta\".");
				}
				my $outputhandle2;
				unless (open($outputhandle2, "> $outputbank.$i." . ($i + 1) . ".fasta.qual")) {
					&errorMessage(__LINE__, "Cannot write \"$outputbank.$i." . ($i + 1) . ".fasta.qual\".");
				}
				for (my $j = $i; $j <= $i + 1; $j ++) {
					my $inputhandle;
					unless (open($inputhandle, "< $outputbank.$j.fasta")) {
						&errorMessage(__LINE__, "Cannot read \"$outputbank.$j.fasta\".");
					}
					while (<$inputhandle>) {
						if (($mergemode eq 'rough' || $mergemode eq 'draft' && scalar(@inputfiles) != 2 || $mergemode eq 'normal' && $step > 1 && scalar(@inputfiles) != 2) && $j == $i && /^>/) {
							$nseqfirst ++;
						}
						s/^>(\d+)(\r?\n?)$/>contig_$j\_$1$2/;
						print($outputhandle1 $_);
					}
					close($inputhandle);
					unless (unlink("$outputbank.$j.fasta")) {
						&errorMessage(__LINE__, "Cannot delete \"$outputbank.$j.fasta\".");
					}
					unless (open($inputhandle, "< $outputbank.$j.fasta.qual")) {
						&errorMessage(__LINE__, "Cannot open \"$outputbank.$j.fasta.qual\".");
					}
					while (<$inputhandle>) {
						s/^>(\d+)(\r?\n?)$/>contig_$j\_$1$2/;
						print($outputhandle2 $_);
					}
					close($inputhandle);
					unless (unlink("$outputbank.$j.fasta.qual")) {
						&errorMessage(__LINE__, "Cannot delete \"$outputbank.$j.fasta.qual\".");
					}
				}
				close($outputhandle1);
				close($outputhandle2);
			}
			# secondary assemble
			if (system("$toAmos -s $outputbank.$i." . ($i + 1) . ".fasta -q $outputbank.$i." . ($i + 1) . ".fasta.qual -gq 30 -bq 10 -o - 2> $devnull | $banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".temp1.bnk -m - 2> $devnull 1> $devnull")) {
				&errorMessage(__LINE__, "Cannot run \"$toAmos -s $outputbank.$i." . ($i + 1) . ".fasta -q $outputbank.$i." . ($i + 1) . ".fasta.qual -gq 30 -bq 10 -o - | $banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".temp1.bnk -m -\".");
			}
			unless (unlink("$outputbank.$i." . ($i + 1) . ".fasta")) {
				&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".fasta\".");
			}
			unless (unlink("$outputbank.$i." . ($i + 1) . ".fasta.qual")) {
				&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".fasta.qual\".");
			}
			if ($minident2 == 1 || $exact) {
				&assembleBank($ungappedoverlap, 0, 0, "$outputbank.$i." . ($i + 1) . ".temp1.bnk", $nseqfirst);
			}
			if ($minident2 < 1) {
				my $tempopt;
				if ($strand) {
					$tempopt .= " --strand=plus";
				}
				if ($maxnsplit) {
					$tempopt .= " --maxnsplits=$maxnsplit";
				}
				if ($ungapped) {
					if (-e "$outputbank.$i." . ($i + 1) . ".temp1.bnk/CTG.ifo") {
						if (system("$assams$tempopt --noeid --assemblemode=ungapped --mergemode=$mergemode --minnseq=$minnseq --minident=1,$minident2,0.7 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.$i." . ($i + 1) . ".temp1.bnk $outputbank.$i." . ($i + 1) . ".temp2.bnk 2> $devnull 1> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --assemblemode=ungapped --mergemode=$mergemode --minnseq=$minnseq --minident=1,$minident2,0.7 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.$i." . ($i + 1) . ".temp1.bnk $outputbank.$i." . ($i + 1) . ".temp2.bnk\".");
						}
						#if (!-e "$outputbank.$i." . ($i + 1) . ".temp2.bnk/CTG.ifo") {
						#	&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --assemblemode=ungapped --mergemode=$mergemode --minnseq=$minnseq --minident=1,$minident2,0.7 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.$i." . ($i + 1) . ".temp1.bnk $outputbank.$i." . ($i + 1) . ".temp2.bnk\".");
						#}
						&deleteBanks("$outputbank.$i." . ($i + 1) . ".temp1.bnk");
						unless (rename("$outputbank.$i." . ($i + 1) . ".temp2.bnk", "$outputbank.$i." . ($i + 1) . ".temp1.bnk")) {
							&errorMessage(__LINE__, "Cannot rename \"$outputbank.$i." . ($i + 1) . ".temp2.bnk\" to \"$outputbank.$i." . ($i + 1) . ".temp1.bnk\".");
						}
					}
					else {
						&assembleBank($ungappedoverlap, (1 - $minident2), 0.6, "$outputbank.$i." . ($i + 1) . ".temp1.bnk", $nseqfirst);
					}
				}
				if ($gapped) {
					if (-e "$outputbank.$i." . ($i + 1) . ".temp1.bnk/CTG.ifo") {
						if (system("$assams$tempopt --noeid --assemblemode=gapped --mergemode=$mergemode --minnseq=$minnseq --minident=1,$minident2,0.7 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.$i." . ($i + 1) . ".temp1.bnk $outputbank.$i." . ($i + 1) . ".temp2.bnk 2> $devnull 1> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --assemblemode=gapped --mergemode=$mergemode --minnseq=$minnseq --minident=1,$minident2,0.7 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.$i." . ($i + 1) . ".temp1.bnk $outputbank.$i." . ($i + 1) . ".temp2.bnk\".");
						}
						#if (!-e "$outputbank.$i." . ($i + 1) . ".temp2.bnk/CTG.ifo") {
						#	&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --assemblemode=gapped --mergemode=$mergemode --minnseq=$minnseq --minident=1,$minident2,0.7 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.$i." . ($i + 1) . ".temp1.bnk $outputbank.$i." . ($i + 1) . ".temp2.bnk\".");
						#}
						&deleteBanks("$outputbank.$i." . ($i + 1) . ".temp1.bnk");
						unless (rename("$outputbank.$i." . ($i + 1) . ".temp2.bnk", "$outputbank.$i." . ($i + 1) . ".temp1.bnk")) {
							&errorMessage(__LINE__, "Cannot rename \"$outputbank.$i." . ($i + 1) . ".temp2.bnk\" to \"$outputbank.$i." . ($i + 1) . ".temp1.bnk\".");
						}
					}
					else {
						&assembleBank($hashoverlap, (1 - $minident2), 0.6, "$outputbank.$i." . ($i + 1) . ".temp1.bnk", $nseqfirst);
					}
				}
			}
			# read assemble results and get contig-reads association
			my %primaryaffiliation; # affiliation of sequence reads to primary contig
			my %secondaryaffiliation; # affiliation of sequence reads to secondary contig
			my %secondarysingletons;
			my %singletons;
			{
				# read 2ndary assemble results
				my $pipehandle;
				my %temp;
				if (-e "$outputbank.$i." . ($i + 1) . ".temp1.bnk/CTG.ifo") {
					unless (open($pipehandle, "$listReadPlacedStatus $outputbank.$i." . ($i + 1) . ".temp1.bnk 2> $devnull |")) {
						&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $outputbank.$i." . ($i + 1) . ".temp1.bnk\".");
					}
					while (<$pipehandle>) {
						s/\r?\n?$//;
						if (my @row = split(/\t/)) {
							if (scalar(@row) == 4 && $row[2] eq 'S' && $row[1] =~ /^contig_(\d+)_(\d+)$/) {
								$secondarysingletons{$1}{$2} = $row[0];
							}
							elsif (scalar(@row) == 4 && $row[2] eq 'S') {
								$singletons{$row[1]} = 1;
							}
							elsif (scalar(@row) == 5 && $row[2] eq 'P' && $row[1] =~ /^contig_\d+_\d+$/) {
								$temp{$&} = $row[4];
							}
							elsif (scalar(@row) == 5 && $row[2] eq 'P') {
								$secondaryaffiliation{$row[1]} = $row[4];
							}
							else {
								&errorMessage(__LINE__, "Invalid assemble results.\nBank: $outputbank.$i." . ($i + 1) . ".temp1.bnk\nSequence: $row[1]\nSecondary contig: $row[4]\n");
							}
						}
					}
					close($pipehandle);
					if ($?) {
						&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $outputbank.$i." . ($i + 1) . ".temp1.bnk\".");
					}
				}
				else {
					unless (open($pipehandle, "$dumpreads -r -e $outputbank.$i." . ($i + 1) . ".temp1.bnk 2> $devnull |")) {
						&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $outputbank.$i." . ($i + 1) . ".temp1.bnk\".");
					}
					while (<$pipehandle>) {
						if (/^>contig_(\d+)_(\d+)/) {
							$secondarysingletons{$1}{$2} = 1;
						}
						elsif (/^>(\S+)/) {
							$singletons{$1} = 1;
						}
					}
					close($pipehandle);
					if ($?) {
						&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $outputbank.$i." . ($i + 1) . ".temp1.bnk\".");
					}
				}
				# read primary assemble results
				for (my $j = $i; $j <= $i + 1; $j ++) {
					if (-e "$inputfiles[$j]/CTG.ifo") {
						unless (open($pipehandle, "$listReadPlacedStatus $inputfiles[$j] 2> $devnull |")) {
							&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $inputfiles[$j]\".");
						}
						while (<$pipehandle>) {
							s/\r?\n?$//;
							if (my @row = split(/\t/)) {
								if (scalar(@row) == 4 && $row[2] eq 'S') {
									if (!exists($secondaryaffiliation{$row[1]}) && !exists($singletons{$row[1]})) {
										&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[$j]\nSequence: $row[1]\n");
									}
								}
								elsif (scalar(@row) == 5 && $row[2] eq 'P' && $temp{"contig_$j\_$row[4]"}) {
									$secondaryaffiliation{$row[1]} = $temp{"contig_$j\_$row[4]"};
								}
								elsif (scalar(@row) == 5 && $row[2] eq 'P') {
									$primaryaffiliation{$j}{$row[1]} = $row[4];
								}
								else {
									&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[$j]\nSequence: $row[1]\nPrimary contig: contig_$j\_$row[4]\nSecondary contig: " . $temp{"contig_$j\_$row[4]"} . "\n");
								}
							}
						}
						close($pipehandle);
						if ($?) {
							&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $inputfiles[$j]\".");
						}
					}
				}
			}
			# get secondary contigs and all reads, and save them to appropreate files
			my $num2ndcontig = 1;
			{
				# make 2ndcontig files
				# sequence
				# secondarycontig
				my $pipehandle;
				my $outputhandle;
				my $inputhandle;
				if (-e "$outputbank.$i." . ($i + 1) . ".temp1.bnk/CTG.ifo") {
					unless (open($pipehandle, "$bank2fasta -b $outputbank.$i." . ($i + 1) . ".temp1.bnk -iid -q $outputbank.$i." . ($i + 1) . ".2ndcontigs.fasta.qual 2> $devnull |")) {
						&errorMessage(__LINE__, "Cannot run \"$bank2fasta -b $outputbank.$i." . ($i + 1) . ".temp1.bnk -iid -q $outputbank.$i." . ($i + 1) . ".2ndcontigs.fasta.qual\".");
					}
					while (<$pipehandle>) {
						if (s/^>(\d+)(\r?\n?)$/>secondarycontig_$num2ndcontig$2/) {
							if ($1 != $num2ndcontig) {
								&errorMessage(__LINE__, "Invalid assemble results.");
							}
							if ($outputhandle) {
								close($outputhandle);
							}
							unless (open($outputhandle, "> $outputbank.$i." . ($i + 1) . ".2ndcontig$num2ndcontig.fasta")) {
								&errorMessage(__LINE__, "Cannot make \"$outputbank.$i." . ($i + 1) . ".2ndcontig$num2ndcontig.fasta\".");
							}
							$num2ndcontig ++;
						}
						print($outputhandle $_);
					}
					if ($outputhandle) {
						close($outputhandle);
					}
					close($pipehandle);
					if ($?) {
						&errorMessage(__LINE__, "Cannot run \"$bank2fasta -b $outputbank.$i." . ($i + 1) . ".temp1.bnk -iid -q $outputbank.$i." . ($i + 1) . ".2ndcontigs.fasta.qual\".");
					}
					# quality
					$num2ndcontig = 1;
					# secondarycontig
					unless (open($inputhandle, "< $outputbank.$i." . ($i + 1) . ".2ndcontigs.fasta.qual")) {
						&errorMessage(__LINE__, "Cannot open \"$outputbank.$i." . ($i + 1) . ".2ndcontigs.fasta.qual\".");
					}
					while (<$inputhandle>) {
						if (s/^>(\d+)(\r?\n?)$/>secondarycontig_$num2ndcontig$2/) {
							if ($1 != $num2ndcontig) {
								&errorMessage(__LINE__, "Invalid assemble results.");
							}
							if ($outputhandle) {
								close($outputhandle);
							}
							unless (open($outputhandle, "> $outputbank.$i." . ($i + 1) . ".2ndcontig$num2ndcontig.fasta.qual")) {
								&errorMessage(__LINE__, "Cannot make \"$outputbank.$i." . ($i + 1) . ".2ndcontig$num2ndcontig.fasta.qual\".");
							}
							$num2ndcontig ++;
						}
						print($outputhandle $_);
					}
					if ($outputhandle) {
						close($outputhandle);
					}
					close($inputhandle);
					unless (unlink("$outputbank.$i." . ($i + 1) . ".2ndcontigs.fasta.qual")) {
						&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".2ndcontigs.fasta.qual\".");
					}
				}
				# make reads files
				for (my $j = $i; $j <= $i + 1; $j ++) {
					unless (open($pipehandle, "$dumpreads -r -e $inputfiles[$j] 2> $devnull |")) {
						&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $inputfiles[$j]\".");
					}
					while (<$pipehandle>) {
						if (/^>(.+)\r?\n?/) {
							my $seqname = $1;
							if ($secondaryaffiliation{$seqname}) {
								if ($outputhandle) {
									close($outputhandle);
								}
								if ($secondaryaffiliation{$seqname} =~ /^\d+$/) {
									unless (open($outputhandle, ">> $outputbank.$i." . ($i + 1) . ".reads$secondaryaffiliation{$seqname}.fasta")) {
										&errorMessage(__LINE__, "Cannot write \"$outputbank.reads$secondaryaffiliation{$seqname}.fasta\".");
									}
								}
								else {
									&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[$j]\nSequence: $seqname\nSecondary contig: $secondaryaffiliation{$seqname}\n");
								}
							}
							elsif ($primaryaffiliation{$j}{$seqname}) {
								if ($outputhandle) {
									close($outputhandle);
								}
								if ($primaryaffiliation{$j}{$seqname} =~ /^\d+$/) {
									unless (open($outputhandle, "> $devnull")) {
										&errorMessage(__LINE__, "Cannot write \"$devnull\".");
									}
								}
								else {
									&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[$j]\nSequence: $seqname\nPrimary contig: $primaryaffiliation{$j}{$seqname}\n");
								}
							}
							elsif ($singletons{$seqname}) {
								if ($outputhandle) {
									close($outputhandle);
								}
								unless (open($outputhandle, ">> $outputbank.$i." . ($i + 1) . ".singletons.fasta")) {
									&errorMessage(__LINE__, "Cannot write \"$outputbank.$i." . ($i + 1) . ".singletons.fasta\".");
								}
							}
							else {
								&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[$j]\nSequence: $seqname\nUnplaced\n");
							}
						}
						print($outputhandle $_);
					}
					if ($outputhandle) {
						close($outputhandle);
					}
					close($pipehandle);
					if ($?) {
						&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $inputfiles[$j]\".");
					}
					unless (open($pipehandle, "$dumpreads -r -e -q $inputfiles[$j] 2> $devnull |")) {
						&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -q $inputfiles[$j]\".");
					}
					while (<$pipehandle>) {
						if (/^>(.+)\r?\n?/) {
							my $seqname = $1;
							if ($secondaryaffiliation{$seqname}) {
								if ($outputhandle) {
									close($outputhandle);
								}
								if ($secondaryaffiliation{$seqname} =~ /^\d+$/) {
									unless (open($outputhandle, ">> $outputbank.$i." . ($i + 1) . ".reads$secondaryaffiliation{$seqname}.fasta.qual")) {
										&errorMessage(__LINE__, "Cannot write \"$outputbank.$i." . ($i + 1) . ".reads$secondaryaffiliation{$seqname}.fasta.qual\".");
									}
								}
								else {
									&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[$j]\nSequence: $seqname\nSecondary contig: $secondaryaffiliation{$seqname}\n");
								}
							}
							elsif ($primaryaffiliation{$j}{$seqname}) {
								if ($outputhandle) {
									close($outputhandle);
								}
								if ($primaryaffiliation{$j}{$seqname} =~ /^\d+$/) {
									unless (open($outputhandle, "> $devnull")) {
										&errorMessage(__LINE__, "Cannot write \"$devnull\".");
									}
								}
								else {
									&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[$j]\nSequence: $seqname\nPrimary contig: $primaryaffiliation{$j}{$seqname}\n");
								}
							}
							elsif ($singletons{$seqname}) {
								if ($outputhandle) {
									close($outputhandle);
								}
								unless (open($outputhandle, ">> $outputbank.$i." . ($i + 1) . ".singletons.fasta.qual")) {
									&errorMessage(__LINE__, "Cannot write \"$outputbank.$i." . ($i + 1) . ".singletons.fasta.qual\".");
								}
							}
							else {
								&errorMessage(__LINE__, "Invalid assemble results.\nBank: $inputfiles[$j]\nSequence: $seqname\nUnplaced\n");
							}
						}
						print($outputhandle $_);
					}
					if ($outputhandle) {
						close($outputhandle);
					}
					close($pipehandle);
					if ($?) {
						&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -q $inputfiles[$j]\".");
					}
				}
				# make secondarysingleton/primarycontig bank
				my $num2ndsingleton = 0;
				for (my $j = $i; $j <= $i + 1; $j ++) {
					if ($secondarysingletons{$j} && $primaryaffiliation{$j}) {
						my $pipehandlein;
						unless (open($pipehandlein, "$bankreport -p -b " . $inputfiles[$j] . " 2> $devnull |")) {
							&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b " . $inputfiles[$j] . "\".");
						}
						my $pipehandleout;
						unless (open($pipehandleout, "| $banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".reads" . ($num2ndcontig + $num2ndsingleton) . ".bnk -m - 2> $devnull 1> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".reads" . ($num2ndcontig + $num2ndsingleton) . ".bnk -m -\".");
						}
						my $temp;
						my $nesting = 0;
						my $contig;
						my $read;
						my $numcontig = 1;
						while (<$pipehandlein>) {
							if ($nesting && /^\}/) {
								$nesting --;
							}
							elsif (/^\{([A-Z]+)/) {
								$nesting ++;
								if ($1 eq 'CTG') {
									$contig ++;
								}
								elsif ($1 eq 'RED') {
									$read ++;
								}
							}
							if ($contig || $read) {
								$temp .= $_;
							}
							if ($temp && $nesting == 0) {
								if ($contig && $temp =~ s/^iid:(\d+)/iid:$numcontig/m && $secondarysingletons{$j}{$1}) {
									$temp =~ s/^eid:\S+\r?\n?//m;
									print($pipehandleout $temp);
									$numcontig ++;
								}
								elsif ($read && $temp =~ /^eid:(\S+)/m && $primaryaffiliation{$j}{$1}) {
									print($pipehandleout $temp);
								}
								undef($temp);
								undef($contig);
								undef($read);
							}
						}
						close($pipehandleout);
						if ($?) {
							&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".reads" . ($num2ndcontig + $num2ndsingleton) . ".bnk -m -\".");
						}
						close($pipehandlein);
						if ($?) {
							&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b " . $inputfiles[$j] . "\".");
						}
						$num2ndsingleton ++;
					}
				}
			}
			# mapping reads to secondary contig
			{
				my $child = 0;
				$| = 1;
				$? = 0;
				for (my $j = 1; $j < $num2ndcontig; $j ++) {
					if (my $pid = fork()) {
						$child ++;
						if ($child >= ($numthreads + 1)) {
							if (wait == -1) {
								$child = 0;
							} else {
								$child --;
							}
						}
						if ($?) {
							&errorMessage(__LINE__);
						}
						next;
					}
					else {
						if (!-e "$outputbank.$i." . ($i + 1) . ".2ndcontig$j.fasta") {
							&errorMessage(__LINE__, "Cannot find \"$outputbank.$i." . ($i + 1) . ".2ndcontig$j.fasta\".")
						}
						if (!-e "$outputbank.$i." . ($i + 1) . ".2ndcontig$j.fasta.qual") {
							&errorMessage(__LINE__, "Cannot find \"$outputbank.$i." . ($i + 1) . ".2ndcontig$j.fasta.qual\".");
						}
						if (system("$toAmos -s $outputbank.$i." . ($i + 1) . ".2ndcontig$j.fasta -q $outputbank.$i." . ($i + 1) . ".2ndcontig$j.fasta.qual -gq 30 -bq 10 -o - 2> $devnull | $banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk -m - 2> $devnull 1> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"$toAmos -s $outputbank.$i." . ($i + 1) . ".2ndcontig$j.fasta -q $outputbank.$i." . ($i + 1) . ".2ndcontig$j.fasta.qual -gq 30 -bq 10 -o - | $banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk -m -\".");
						}
						unless (unlink("$outputbank.$i." . ($i + 1) . ".2ndcontig$j.fasta")) {
							&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".2ndcontig$j.fasta\"");
						}
						unless (unlink("$outputbank.$i." . ($i + 1) . ".2ndcontig$j.fasta.qual")) {
							&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".2ndcontig$j.fasta.qual\"");
						}
						&cleanBank("$outputbank.$i." . ($i + 1) . ".reads$j.bnk", 'RED');
						if (!-e "$outputbank.$i." . ($i + 1) . ".reads$j.fasta") {
							&errorMessage(__LINE__, "Cannot find \"$outputbank.$i." . ($i + 1) . ".reads$j.fasta\".")
						}
						if (!-e "$outputbank.$i." . ($i + 1) . ".reads$j.fasta.qual") {
							&errorMessage(__LINE__, "Cannot find \"$outputbank.$i." . ($i + 1) . ".reads$j.fasta.qual\".");
						}
						if (system("$toAmos -s $outputbank.$i." . ($i + 1) . ".reads$j.fasta -q $outputbank.$i." . ($i + 1) . ".reads$j.fasta.qual -gq 30 -bq 10 -id 2 -o - 2> $devnull | $banktransact -z -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk -m - 2> $devnull 1> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"$toAmos -s $outputbank.$i." . ($i + 1) . ".reads$j.fasta -q $outputbank.$i." . ($i + 1) . ".reads$j.fasta.qual -gq 30 -bq 10 -id 2 -o - | $banktransact -z -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk -m -\".");
						}
						if (!$ungapped && !$gapped) {
							if (system("OMP_NUM_THREADS=1 $ungappedoverlap$strand -B -r 1 -o $minovllen -m 0 -x 0 -v 0 $outputbank.$i." . ($i + 1) . ".reads$j.bnk 2> $devnull 1> $devnull")) {
								&errorMessage(__LINE__, "Cannot run \"OMP_NUM_THREADS=1 $ungappedoverlap$strand -B -r 1 -o $minovllen -m 0 -x 0 -v 0 $outputbank.$i." . ($i + 1) . ".reads$j.bnk\".");
							}
						}
						if (!$gapped) {
							if (system("OMP_NUM_THREADS=1 $ungappedoverlap$strand -B -r 1 -o $minovllen -m 0 -x 0.5 -v 0 $outputbank.$i." . ($i + 1) . ".reads$j.bnk 2> $devnull 1> $devnull")) {
								&errorMessage(__LINE__, "Cannot run \"OMP_NUM_THREADS=1 $ungappedoverlap$strand -B -r 1 -o $minovllen -m 0 -x 0.5 -v 0 $outputbank.$i." . ($i + 1) . ".reads$j.bnk\".");
							}
						}
						else {
							if (system("OMP_NUM_THREADS=1 $hashoverlap$strand -B -r 1 -o $minovllen -m 0 -x 0.5 -v 0 $outputbank.$i." . ($i + 1) . ".reads$j.bnk 2> $devnull 1> $devnull")) {
								&errorMessage(__LINE__, "Cannot run \"OMP_NUM_THREADS=1 $hashoverlap$strand -B -r 1 -o $minovllen -m 0 -x 0.5 -v 0 $outputbank.$i." . ($i + 1) . ".reads$j.bnk\".");
							}
						}
						if (system("$tigger -v 0 -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk 2> $devnull 1> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"$tigger -v 0 -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk\".");
						}
						&cleanBank("$outputbank.$i." . ($i + 1) . ".reads$j.bnk", 'LAY', 'RED');
						if (system("$bankreport -p -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk > $outputbank.$i." . ($i + 1) . ".reads$j.afg 2> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk > $outputbank.$i." . ($i + 1) . ".reads$j.afg\".");
						}
						&deleteBanks("$outputbank.$i." . ($i + 1) . ".reads$j.bnk");
						{
							my $inputhandle;
							unless (open($inputhandle, "< $outputbank.$i." . ($i + 1) . ".reads$j.afg")) {
								&errorMessage(__LINE__, "Cannot read \"$outputbank.$i." . ($i + 1) . ".reads$j.afg\".");
							}
							my $pipehandleout;
							unless (open($pipehandleout, "| $banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk -m - 1> $devnull 2> $devnull")) {
								&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk -m -\".");
							}
							my $temp;
							my $nesting = 0;
							my $tle;
							my $read;
							while (<$inputhandle>) {
								if ($nesting && /^\}/) {
									$nesting --;
								}
								elsif (/^\{(TLE|RED)/) {
									$nesting ++;
									if ($1 eq 'TLE') {
										$tle ++;
									}
									elsif ($1 eq 'RED') {
										$read ++;
									}
								}
								elsif ($nesting == 0) {
									print($pipehandleout $_);
								}
								if ($tle || $read) {
									$temp .= $_;
								}
								if ($temp && $nesting == 0) {
									if ($tle && $temp !~ /^src:1\r?\n?$/m) {
										print($pipehandleout $temp);
									}
									elsif ($read && $temp !~ /^iid:1\r?\n?$/m) {
										print($pipehandleout $temp);
									}
									undef($temp);
									undef($tle);
									undef($read);
								}
							}
							close($pipehandleout);
							if ($?) {
								&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk -m -\".");
							}
							close($inputhandle);
							unless (unlink("$outputbank.$i." . ($i + 1) . ".reads$j.afg")) {
								&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".reads$j.afg\"");
							}
						}
						# if mapping failed
						if (system("$makeconsensus -b -B -o $minovllen -e " . ((1 - $minident3) * 2) . " -w $alnwiggle -v 0 -x $outputbank.$i." . ($i + 1) . ".reads$j.txt $outputbank.$i." . ($i + 1) . ".reads$j.bnk 2> $devnull 1> $devnull")) {
							print(STDERR "Mapping failed at \"$outputbank.$i." . ($i + 1) . ".reads$j.bnk\". Assembly will be retried.\n");
							&deleteBanks("$outputbank.$i." . ($i + 1) . ".reads$j.bnk");
						}
						# if mapping succeeded
						else {
							my %expels;
							my $inputhandle;
							unless (open($inputhandle, "< $outputbank.$i." . ($i + 1) . ".reads$j.txt")) {
								&errorMessage(__LINE__, "Cannot read \"$outputbank.$i." . ($i + 1) . ".reads$j.txt\".");
							}
							while (<$inputhandle>) {
								if (/^(\d+)\s+\d+/) {
									$expels{$1} = 1;
								}
							}
							close($inputhandle);
							# if there are expelled sequences, make new bank
							if (%expels) {
								my $outputhandle;
								unless (open($outputhandle, "> $outputbank.$i." . ($i + 1) . ".reads$j.expels.iid")) {
									&errorMessage(__LINE__, "Cannot make \"$outputbank.$i." . ($i + 1) . ".reads$j.expels.iid\".");
								}
								foreach my $iid (keys(%expels)) {
									print($outputhandle "$iid\n");
								}
								close($outputhandle);
								if (system("$dumpreads -r -e -I $outputbank.$i." . ($i + 1) . ".reads$j.expels.iid $outputbank.$i." . ($i + 1) . ".reads$j.bnk > $outputbank.$i." . ($i + 1) . ".reads$j.expels.fasta 2> $devnull")) {
									&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -I $outputbank.$i." . ($i + 1) . ".reads$j.expels.iid $outputbank.$i." . ($i + 1) . ".reads$j.bnk > $outputbank.$i." . ($i + 1) . ".reads$j.expels.fasta\".");
								}
								if (system("$dumpreads -r -e -q -I $outputbank.$i." . ($i + 1) . ".reads$j.expels.iid $outputbank.$i." . ($i + 1) . ".reads$j.bnk > $outputbank.$i." . ($i + 1) . ".reads$j.expels.fasta.qual 2> $devnull")) {
									&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -q -I $outputbank.$i." . ($i + 1) . ".reads$j.expels.iid $outputbank.$i." . ($i + 1) . ".reads$j.bnk > $outputbank.$i." . ($i + 1) . ".reads$j.expels.fasta.qual\".");
								}
								unless (unlink("$outputbank.$i." . ($i + 1) . ".reads$j.expels.iid")) {
									&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".reads$j.expels.iid\".");
								}
								my $pipehandlein;
								unless (open($pipehandlein, "$bankreport -p -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk 2> $devnull |")) {
									&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk\".");
								}
								my $pipehandleout;
								unless (open($pipehandleout, "| $banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".reads$j.temp.bnk -m - 1> $devnull 2> $devnull")) {
									&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".reads$j.temp.bnk -m -\".");
								}
								my $temp;
								my $nesting = 0;
								my $contig;
								my $read;
								while (<$pipehandlein>) {
									if ($nesting && /^\}/) {
										$nesting --;
									}
									elsif (/^\{([A-Z]+)/) {
										$nesting ++;
										if ($1 eq 'CTG') {
											$contig ++;
										}
										elsif ($1 eq 'RED') {
											$read ++;
										}
									}
									#elsif ($nesting == 0) {
									#	print($pipehandleout $_);
									#}
									if ($contig || $read) {
										$temp .= $_;
									}
									if ($temp && $nesting == 0) {
										if ($contig || ($read && $temp =~ /^iid:(\d+)\r?\n?$/m && !$expels{$1})) {
											print($pipehandleout $temp);
										}
										undef($temp);
										undef($contig);
										undef($read);
									}
								}
								close($pipehandleout);
								if ($?) {
									&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".reads$j.temp.bnk -m -\".");
								}
								close($pipehandlein);
								if ($?) {
									&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk\".");
								}
								&deleteBanks("$outputbank.$i." . ($i + 1) . ".reads$j.bnk");
								unless (rename("$outputbank.$i." . ($i + 1) . ".reads$j.temp.bnk", "$outputbank.$i." . ($i + 1) . ".reads$j.bnk")) {
									&errorMessage(__LINE__, "Cannot rename \"$outputbank.$i." . ($i + 1) . ".reads$j.temp.bnk\" to \"$outputbank.$i." . ($i + 1) . ".reads$j.bnk\".");
								}
							}
						}
						exit;
					}
				}
			}
			# join
			while (wait != -1) {
				if ($?) {
					&errorMessage(__LINE__, 'Cannot run merging pipeline correctly.');
				}
			}
			my $tempopt;
			if ($strand) {
				$tempopt .= " --strand=plus";
			}
			if ($maxnsplit) {
				$tempopt .= " --maxnsplits=$maxnsplit";
			}
			if ($exact && $ungapped && $gapped) {
				$tempopt .= " --assemblemode=all";
			}
			elsif ($exact && $ungapped) {
				$tempopt .= " --assemblemode=e+u";
			}
			elsif ($exact && $gapped) {
				$tempopt .= " --assemblemode=e+g";
			}
			elsif ($ungapped && $gapped) {
				$tempopt .= " --assemblemode=u+g";
			}
			elsif ($exact) {
				$tempopt .= " --assemblemode=exact";
			}
			elsif ($ungapped) {
				$tempopt .= " --assemblemode=ungapped";
			}
			elsif ($gapped) {
				$tempopt .= " --assemblemode=gapped";
			}
			if ($minident3 == 1) {
				$tempopt .= " --mergemode=rough";
			}
			else {
				$tempopt .= " --mergemode=$mergemode";
			}
			# retry assemble if mapping failed
			for (my $j = 1; $j < $num2ndcontig; $j ++) {
				if (!-e "$outputbank.$i." . ($i + 1) . ".reads$j.bnk" && -e "$outputbank.$i." . ($i + 1) . ".reads$j.fasta" && -e "$outputbank.$i." . ($i + 1) . ".reads$j.fasta.qual") {
					print(STDERR "Retrying assembly of \"$outputbank.$i." . ($i + 1) . ".reads$j.bnk\"...");
					if (system("$assams$tempopt --noeid --minnseq=$minnseq --minident=$minident3 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.$i." . ($i + 1) . ".reads$j.fasta $outputbank.$i." . ($i + 1) . ".reads$j.bnk 2> $devnull 1> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --minnseq=$minnseq --minident=$minident3 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.$i." . ($i + 1) . ".reads$j.fasta $outputbank.$i." . ($i + 1) . ".reads$j.bnk\".");
					}
					#if (!-e "$outputbank.$i." . ($i + 1) . ".reads$j.bnk/CTG.ifo") {
					#	&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --minnseq=$minnseq --minident=$minident3 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.$i." . ($i + 1) . ".reads$j.fasta $outputbank.$i." . ($i + 1) . ".reads$j.bnk\".");
					#}
					#if (!-e "$outputbank.$i." . ($i + 1) . ".reads$j.bnk/CTG.ifo" && !-e "$outputbank.$i." . ($i + 1) . ".reads$j.bnk/LAY.ifo" && !-e "$outputbank.$i." . ($i + 1) . ".reads$j.bnk/OVL.ifo") {
					#	if (system("OMP_NUM_THREADS=$numthreads $hashoverlap$strand -B -o $minovllen -m $maxnhits -x " . (1 - $minident3) . " -v 0 $outputbank.$i." . ($i + 1) . ".reads$j.bnk 2> $devnull 1> $devnull")) {
					#		&errorMessage(__LINE__, "Cannot run \"OMP_NUM_THREADS=$numthreads $hashoverlap$strand -B -o $minovllen -m $maxnhits -x " . (1 - $minident3) . " -v 0 $outputbank.$i." . ($i + 1) . ".reads$j.bnk 2> $devnull 1> $devnull\".");
					#	}
					#}
					#if (!-e "$outputbank.$i." . ($i + 1) . ".reads$j.bnk/CTG.ifo" && !-e "$outputbank.$i." . ($i + 1) . ".reads$j.bnk/LAY.ifo") {
					#	if (system("$tigger -v 0 -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk 2> $devnull 1> $devnull")) {
					#		&errorMessage(__LINE__, "Cannot run \"$tigger -v 0 -b $outputbank.$i." . ($i + 1) . ".reads$j.bnk\".");
					#	}
					#}
					#if (!-e "$outputbank.$i." . ($i + 1) . ".reads$j.bnk/CTG.ifo") {
					#	if (system("$makeconsensus -b -B -o $minovllen -e " . ((1 - $minident3) * 2) . " -w $alnwiggle -v 0 $outputbank.$i." . ($i + 1) . ".reads$j.bnk 2> $devnull 1> $devnull")) {
					#		&errorMessage(__LINE__, "Cannot run \"$makeconsensus -b -B -o $minovllen -e " . ((1 - $minident3) * 2) . " -w $alnwiggle -v 0 $outputbank.$i." . ($i + 1) . ".reads$j.bnk\".");
					#	}
					#}
					print(STDERR "OK.\n");
				}
				elsif (!-e "$outputbank.$i." . ($i + 1) . ".reads$j.bnk" || !-e "$outputbank.$i." . ($i + 1) . ".reads$j.bnk/CTG.ifo" || !-e "$outputbank.$i." . ($i + 1) . ".reads$j.bnk/RED.ifo") {
					&errorMessage(__LINE__, "\"$outputbank.$i." . ($i + 1) . ".reads$j.bnk\" is invalid.");
				}
				if (-e "$outputbank.$i." . ($i + 1) . ".reads$j.txt") {
					unless (unlink("$outputbank.$i." . ($i + 1) . ".reads$j.txt")) {
						&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".reads$j.txt\"");
					}
				}
				if (-e "$outputbank.$i." . ($i + 1) . ".reads$j.fasta") {
					unless (unlink("$outputbank.$i." . ($i + 1) . ".reads$j.fasta")) {
						&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".reads$j.fasta\"");
					}
				}
				if (-e "$outputbank.$i." . ($i + 1) . ".reads$j.fasta.qual") {
					unless (unlink("$outputbank.$i." . ($i + 1) . ".reads$j.fasta.qual")) {
						&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".reads$j.fasta.qual\"");
					}
				}
				# assemble expelled sequences
				if (-e "$outputbank.$i." . ($i + 1) . ".reads$j.expels.fasta" && -e "$outputbank.$i." . ($i + 1) . ".reads$j.expels.fasta.qual") {
					print(STDERR "Assembling expelled sequences from \"$outputbank.$i." . ($i + 1) . ".reads$j.bnk\"...");
					if (system("$assams$tempopt --noeid --minnseq=$minnseq --minident=$minident3 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.$i." . ($i + 1) . ".reads$j.expels.fasta $outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk 2> $devnull 1> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --minnseq=$minnseq --minident=$minident3 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.$i." . ($i + 1) . ".reads$j.expels.fasta $outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk\".");
					}
					#if (!-e "$outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk/CTG.ifo") {
					#	&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --minnseq=$minnseq --minident=$minident3 --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $outputbank.$i." . ($i + 1) . ".reads$j.expels.fasta $outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk\".");
					#}
					#if (!-e "$outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk/CTG.ifo" && !-e "$outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk/LAY.ifo" && !-e "$outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk/OVL.ifo") {
					#	if (system("OMP_NUM_THREADS=$numthreads $hashoverlap$strand -B -o $minovllen -m $maxnhits -x " . (1 - $minident3) . " -v 0 $outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk 2> $devnull 1> $devnull")) {
					#		&errorMessage(__LINE__, "Cannot run \"OMP_NUM_THREADS=$numthreads $hashoverlap$strand -B -o $minovllen -m $maxnhits -x " . (1 - $minident3) . " -v 0 $outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk 2> $devnull 1> $devnull\".");
					#	}
					#}
					#if (!-e "$outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk/CTG.ifo" && !-e "$outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk/LAY.ifo") {
					#	if (system("$tigger -v 0 -b $outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk 2> $devnull 1> $devnull")) {
					#		&errorMessage(__LINE__, "Cannot run \"$tigger -v 0 -b $outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk\".");
					#	}
					#}
					#if (!-e "$outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk/CTG.ifo") {
					#	if (system("$makeconsensus -b -B -o $minovllen -e " . ((1 - $minident3) * 2) . " -w $alnwiggle -v 0 $outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk 2> $devnull 1> $devnull")) {
					#		&errorMessage(__LINE__, "Cannot run \"$makeconsensus -b -B -o $minovllen -e " . ((1 - $minident3) * 2) . " -w $alnwiggle -v 0 $outputbank.$i." . ($i + 1) . ".reads$j.expels.bnk\".");
					#	}
					#}
					unless (unlink("$outputbank.$i." . ($i + 1) . ".reads$j.expels.fasta")) {
						&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".reads$j.expels.fasta\"");
					}
					unless (unlink("$outputbank.$i." . ($i + 1) . ".reads$j.expels.fasta.qual")) {
						&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".reads$j.expels.fasta.qual\"");
					}
					print(STDERR "OK.\n");
				}
			}
			# make combined bank
			{
				if (-e "$outputbank.$i." . ($i + 1) . ".reads1.bnk") {
					my @readsbanks = glob("$outputbank.$i." . ($i + 1) . ".reads*.bnk");
					if (scalar(@readsbanks) > 1) {
						my ($placed, $unplaced) = &sortBanks(@readsbanks);
						@readsbanks = @{$placed};
						foreach my $readsbank (@{$unplaced}) {
							if (system("$dumpreads -r -e $readsbank >> $outputbank.$i." . ($i + 1) . ".singletons.fasta 2> $devnull")) {
								&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e $readsbank >> $outputbank.$i." . ($i + 1) . ".singletons.fasta\".");
							}
							if (system("$dumpreads -r -e -q $readsbank >> $outputbank.$i." . ($i + 1) . ".singletons.fasta.qual 2> $devnull")) {
								&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -q $readsbank >> $outputbank.$i." . ($i + 1) . ".singletons.fasta.qual\".");
							}
							&deleteBanks($readsbank);
						}
						my $tempbank = shift(@readsbanks);
						while (@readsbanks) {
							my @tempbanks = splice(@readsbanks, 0, 20);
							if (system("$bankcombine $outputbank.$i." . ($i + 1) . ".temp2.bnk $tempbank @tempbanks")) {
								&errorMessage(__LINE__, "Cannot run \"$bankcombine $outputbank.$i." . ($i + 1) . ".temp2.bnk $tempbank @tempbanks\".");
							}
							&deleteBanks($tempbank, @tempbanks);
							if (@readsbanks) {
								unless (rename("$outputbank.$i." . ($i + 1) . ".temp2.bnk", $tempbank)) {
									&errorMessage(__LINE__, "Cannot rename \"$outputbank.$i." . ($i + 1) . ".temp2.bnk\" to \"$tempbank\".");
								}
							}
						}
					}
					else {
						unless (rename("$outputbank.$i." . ($i + 1) . ".reads1.bnk", "$outputbank.$i." . ($i + 1) . ".temp2.bnk")) {
							&errorMessage(__LINE__, "Cannot rename \"$outputbank.$i." . ($i + 1) . ".reads1.bnk\" to \"$outputbank.$i." . ($i + 1) . ".temp2.bnk\".");
						}
					}
				}
				if (-e "$outputbank.$i." . ($i + 1) . ".temp2.bnk") {
					my $maxiid = 1;
					my $pipehandlein;
					unless (open($pipehandlein, "$bankreport -p -b $outputbank.$i." . ($i + 1) . ".temp2.bnk 2> $devnull |")) {
						&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $outputbank.$i." . ($i + 1) . ".temp2.bnk\".");
					}
					my $pipehandleout;
					unless (open($pipehandleout, "| $banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".bnk -m - 1> $devnull 2> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".bnk -m -\".");
					}
					my $temp;
					my $nesting = 0;
					my $contig;
					my $read;
					while (<$pipehandlein>) {
						if (/^iid:(\d+)/ && $1 > $maxiid) {
							$maxiid = $1;
						}
						if ($nesting && /^\}/) {
							$nesting --;
						}
						elsif (/^\{([A-Z]+)/) {
							$nesting ++;
							if ($1 eq 'CTG') {
								$contig ++;
							}
							elsif ($1 eq 'RED') {
								$read ++;
							}
						}
						if ($contig || $read) {
							$temp .= $_;
						}
						if ($temp && $nesting == 0) {
							if ($contig) {
								$temp =~ s/^eid:\S+\r?\n?//m;
								unless ($noeid) {
									$temp =~ s/^iid:(\d+)/$&\neid:contig_$1/m;
								}
							}
							print($pipehandleout $temp);
							undef($temp);
							undef($contig);
							undef($read);
						}
					}
					close($pipehandleout);
					if ($?) {
						&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".bnk -m -\".");
					}
					close($pipehandlein);
					if ($?) {
						&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $outputbank.$i." . ($i + 1) . ".temp2.bnk\".");
					}
					# delete temporary bank
					&deleteBanks("$outputbank.$i." . ($i + 1) . ".temp2.bnk");
					&cleanBank("$outputbank.$i." . ($i + 1) . ".bnk", 'CTG', 'RED');
					$maxiid ++;
					if (-e "$outputbank.$i." . ($i + 1) . ".singletons.fasta" && -e "$outputbank.$i." . ($i + 1) . ".singletons.fasta.qual") {
						if (system("$toAmos -s $outputbank.$i." . ($i + 1) . ".singletons.fasta -q $outputbank.$i." . ($i + 1) . ".singletons.fasta.qual -gq 30 -bq 10 -id $maxiid -o - 2> $devnull | $banktransact -z -b $outputbank.$i." . ($i + 1) . ".bnk -m - 2> $devnull 1> $devnull")) {
							&errorMessage(__LINE__, "Cannot run \"$toAmos -s $outputbank.$i." . ($i + 1) . ".singletons.fasta -q $outputbank.$i." . ($i + 1) . ".singletons.fasta.qual -gq 30 -bq 10 -id $maxiid -o - | $banktransact -z -b $outputbank.$i." . ($i + 1) . ".bnk -m -\".");
						}
						unless (unlink("$outputbank.$i." . ($i + 1) . ".singletons.fasta")) {
							&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".singletons.fasta\".");
						}
						unless (unlink("$outputbank.$i." . ($i + 1) . ".singletons.fasta.qual")) {
							&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".singletons.fasta.qual\".");
						}
					}
				}
				elsif (-e "$outputbank.$i." . ($i + 1) . ".singletons.fasta" && -e "$outputbank.$i." . ($i + 1) . ".singletons.fasta.qual") {
					if (system("$toAmos -s $outputbank.$i." . ($i + 1) . ".singletons.fasta -q $outputbank.$i." . ($i + 1) . ".singletons.fasta.qual -gq 30 -bq 10 -o - 2> $devnull | $banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".bnk -m - 2> $devnull 1> $devnull")) {
						&errorMessage(__LINE__, "Cannot run \"$toAmos -s $outputbank.$i." . ($i + 1) . ".singletons.fasta -q $outputbank.$i." . ($i + 1) . ".singletons.fasta.qual -gq 30 -bq 10 -o - | $banktransact -c -f -z -b $outputbank.$i." . ($i + 1) . ".bnk -m -\".");
					}
					unless (unlink("$outputbank.$i." . ($i + 1) . ".singletons.fasta")) {
						&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".singletons.fasta\".");
					}
					unless (unlink("$outputbank.$i." . ($i + 1) . ".singletons.fasta.qual")) {
						&errorMessage(__LINE__, "Cannot delete \"$outputbank.$i." . ($i + 1) . ".singletons.fasta.qual\".");
					}
				}
				else {
					if (system("$bankcombine $outputbank.$i." . ($i + 1) . '.bnk ' . $inputfiles[$i] . ' ' . $inputfiles[($i + 1)])) {
						&errorMessage(__LINE__, "Cannot run \"$bankcombine $outputbank.$i." . ($i + 1) . '.bnk ' . $inputfiles[$i] . ' ' . $inputfiles[($i + 1)] . "\".");
					}
				}
			}
			# delete temporary bank
			for (my $j = $i; $j <= $i + 1; $j ++) {
				if ($inputfiles[$j] =~ /^$outputbank\.\d+\.bnk$/) {
					&deleteBanks($inputfiles[$j]);
				}
			}
			# delete temporary bank
			&deleteBanks("$outputbank.$i." . ($i + 1) . ".temp1.bnk");
		}
		if (scalar(@inputfiles) > 2) {
			print(STDERR "Preparing next step...\n");
		}
		else {
			print(STDERR "Finishing last step...\n");
		}
		# renew inputfiles
		my @newinputfiles;
		if (scalar(@inputfiles) % 2 == 1) {
			my $newbank = "$outputbank." . scalar(@newinputfiles) . ".bnk";
			if ($newbank ne $inputfiles[-1]) {
				if (-e $newbank) {
					&errorMessage(__LINE__, "\"$newbank\" already exists.");
				}
				elsif ($inputfiles[-1] =~ /^$outputbank\.\d+\.bnk$/) {
					unless (rename($inputfiles[-1], $newbank)) {
						&errorMessage(__LINE__, "Cannot rename \"$inputfiles[-1]\" to \"$newbank\".");
					}
				}
				else {
					unless (dircopy($inputfiles[-1], $newbank)) {
						&errorMessage(__LINE__, "Cannot copy \"$inputfiles[-1]\" to \"$newbank\".");
					}
				}
			}
			push(@newinputfiles, $newbank);
		}
		for (my $i = 0; $i < scalar(@inputfiles) - 1; $i += 2) {
			my $newbank = "$outputbank." . scalar(@newinputfiles) . ".bnk";
			if (-e $newbank) {
				&errorMessage(__LINE__, "\"$newbank\" already exists.");
			}
			unless (rename("$outputbank.$i." . ($i + 1) . ".bnk", $newbank)) {
				&errorMessage(__LINE__, "Cannot rename \"$outputbank.$i." . ($i + 1) . ".bnk\" to \"$newbank\".");
			}
			push(@newinputfiles, $newbank);
		}
		if (scalar(@inputfiles) % 2 == 1 && ((scalar(@inputfiles) - 1) / 2) + 1 != scalar(@newinputfiles) || scalar(@inputfiles) % 2 == 0 && scalar(@inputfiles) / 2 != scalar(@newinputfiles)) {
			&errorMessage(__LINE__, "The number of banks in next step is inconsistent with the number of banks in current step.");
		}
		@inputfiles = @newinputfiles;
		$step ++;
	}
	unless (rename($inputfiles[0], $outputbank)) {
		&errorMessage(__LINE__, "Cannot rename \"$inputfiles[0]\" to \"$outputbank\".");
	}
	&cleanBank($outputbank, 'CTG', 'RED');
	print(STDERR "done.\n\n");
}

sub sortBanks {
	my @placed;
	my @unplaced;
	foreach my $readsbank (@_) {
		my $nplaced = 0;
		my $pipehandle;
		# read assemble results
		if (-e "$readsbank/CTG.ifo") {
			unless (open($pipehandle, "$listReadPlacedStatus $readsbank 2> $devnull |")) {
				&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $readsbank\".");
			}
			while (<$pipehandle>) {
				s/\r?\n?$//;
				if (my @row = split(/\t/)) {
					if (scalar(@row) == 4 && $row[2] eq 'S') {
						next;
					}
					elsif (scalar(@row) == 5 && $row[2] eq 'P') {
						$nplaced ++;
					}
					else {
						&errorMessage(__LINE__, "Invalid assemble results.\nBank: $readsbank\nSequence: $row[1]\n");
					}
				}
			}
			close($pipehandle);
			if ($?) {
				&errorMessage(__LINE__, "Cannot run \"$listReadPlacedStatus $readsbank\".");
			}
		}
		if ($nplaced) {
			push(@placed, $readsbank);
			&cleanBank($readsbank, 'CTG', 'RED');
		}
		else {
			push(@unplaced, $readsbank);
			&cleanBank($readsbank, 'RED');
		}
	}
	return(\@placed, \@unplaced);
}

sub assembleBank {
	my $overlapper = shift(@_);
	my $maxovlerr = shift(@_);
	my $maxconerr = shift(@_);
	my $inputbank = shift(@_);
	my $nseqfirst = shift(@_);
	if ($nseqfirst) {
		$overlapper .= " -r $nseqfirst";
	}
	# overlap
	if (!-e "$inputbank/OVL.ifo" && !-e "$inputbank/LAY.ifo" && !-e "$inputbank/CTG.ifo") {
		if (system("OMP_NUM_THREADS=$numthreads $overlapper$strand -B -o $minovllen -m $maxnhits -x $maxovlerr -v 0 $inputbank 2> $devnull 1> $devnull")) {
			&errorMessage(__LINE__, "Cannot run \"OMP_NUM_THREADS=$numthreads $overlapper$strand -B -o $minovllen -m $maxnhits -x $maxovlerr -v 0 $inputbank\".");
		}
	}
	# layout
	if (!-e "$inputbank/LAY.ifo" && !-e "$inputbank/CTG.ifo") {
		if (system("$tigger -v 0 -b $inputbank 2> $devnull 1> $devnull")) {
			&errorMessage(__LINE__, "Cannot run \"$tigger -v 0 -b $inputbank\".");
		}
	}
	# consensus
	if (!-e "$inputbank/CTG.ifo") {
		if (system("$makeconsensus -b -B -o $minovllen -e $maxconerr -w $alnwiggle -v 0 -x $inputbank.txt $inputbank 2> $devnull 1> $devnull")) {
			&errorMessage(__LINE__, "Cannot run \"$makeconsensus -b -B -o $minovllen -e $maxconerr -w $alnwiggle -v 0 -x $inputbank.txt $inputbank\".");
		}
	}
	if ($maxovlerr > 0 && $maxconerr == $maxovlerr * 2) {
		# read expel file
		my %expels;
		my $inputhandle;
		unless (open($inputhandle, "< $inputbank.txt")) {
			&errorMessage(__LINE__, "Cannot read \"$inputbank.txt\".");
		}
		while (<$inputhandle>) {
			if (/^(\d+)\s+\d+/) {
				$expels{$1} = 1;
			}
		}
		close($inputhandle);
		if (%expels) {
			# if there are expelled sequences, make new bank
			my $outputhandle;
			unless (open($outputhandle, "> $inputbank.expels.iid")) {
				&errorMessage(__LINE__, "Cannot make \"$inputbank.expels.iid\".");
			}
			foreach my $iid (keys(%expels)) {
				print($outputhandle "$iid\n");
			}
			close($outputhandle);
			if (system("$dumpreads -r -e -I $inputbank.expels.iid $inputbank > $inputbank.expels.fasta 2> $devnull")) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -I $inputbank.expels.iid $inputbank > $inputbank.expels.fasta\".");
			}
			if (system("$dumpreads -r -e -q -I $inputbank.expels.iid $inputbank > $inputbank.expels.fasta.qual 2> $devnull")) {
				&errorMessage(__LINE__, "Cannot run \"$dumpreads -r -e -q -I $inputbank.expels.iid $inputbank > $inputbank.expels.fasta.qual\".");
			}
			unless (unlink("$inputbank.expels.iid")) {
				&errorMessage(__LINE__, "Cannot delete \"$inputbank.expels.iid\".");
			}
			my $pipehandlein;
			unless (open($pipehandlein, "$bankreport -p -b $inputbank 2> $devnull |")) {
				&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $inputbank\".");
			}
			my $pipehandleout;
			unless (open($pipehandleout, "| $banktransact -c -f -z -b $inputbank.temp.bnk -m - 1> $devnull 2> $devnull")) {
				&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $inputbank.temp.bnk -m -\".");
			}
			my $temp;
			my $nesting = 0;
			my $contig;
			my $read;
			while (<$pipehandlein>) {
				if ($nesting && /^\}/) {
					$nesting --;
				}
				elsif (/^\{([A-Z]+)/) {
					$nesting ++;
					if ($1 eq 'CTG') {
						$contig ++;
					}
					elsif ($1 eq 'RED') {
						$read ++;
					}
				}
				#elsif ($nesting == 0) {
				#	print($pipehandleout $_);
				#}
				if ($contig || $read) {
					$temp .= $_;
				}
				if ($temp && $nesting == 0) {
					if ($contig || ($read && $temp =~ /^iid:(\d+)\r?\n?$/m && !$expels{$1})) {
						print($pipehandleout $temp);
					}
					undef($temp);
					undef($contig);
					undef($read);
				}
			}
			close($pipehandleout);
			if ($?) {
				&errorMessage(__LINE__, "Cannot run \"$banktransact -c -f -z -b $inputbank.temp.bnk -m -\".");
			}
			close($pipehandlein);
			if ($?) {
				&errorMessage(__LINE__, "Cannot run \"$bankreport -p -b $inputbank\".");
			}
			&deleteBanks($inputbank);
			unless (rename("$inputbank.temp.bnk", $inputbank)) {
				&errorMessage(__LINE__, "Cannot rename \"$inputbank.temp.bnk\" to \"$inputbank\".");
			}
			# re-assemble expelled sequences
			my $tempopt;
			if ($strand) {
				$tempopt .= " --strand=plus";
			}
			if ($maxnsplit) {
				$tempopt .= " --maxnsplits=$maxnsplit";
			}
			if ($exact && $ungapped && $gapped) {
				$tempopt .= " --assemblemode=all";
			}
			elsif ($exact && $ungapped) {
				$tempopt .= " --assemblemode=e+u";
			}
			elsif ($exact && $gapped) {
				$tempopt .= " --assemblemode=e+g";
			}
			elsif ($ungapped && $gapped) {
				$tempopt .= " --assemblemode=u+g";
			}
			elsif ($exact) {
				$tempopt .= " --assemblemode=exact";
			}
			elsif ($ungapped) {
				$tempopt .= " --assemblemode=ungapped";
			}
			elsif ($gapped) {
				$tempopt .= " --assemblemode=gapped";
			}
			if ($maxovlerr == 0) {
				$tempopt .= " --mergemode=rough";
			}
			else {
				$tempopt .= " --mergemode=$mergemode";
			}
			if (system("$assams$tempopt --noeid --minnseq=$minnseq --minident=" . (1 - $maxovlerr) . " --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $inputbank.expels.fasta $inputbank.expels.bnk 2> $devnull 1> $devnull")) {
				&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --minnseq=$minnseq --minident=" . (1 - $maxovlerr) . " --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $inputbank.expels.fasta $inputbank.expels.bnk\".");
			}
			#if (!-e "$inputbank.expels.bnk/CTG.ifo") {
			#	&errorMessage(__LINE__, "Cannot run \"$assams$tempopt --noeid --minnseq=$minnseq --minident=" . (1 - $maxovlerr) . " --minovllen=$minovllen --maxnhits=$maxnhits --alnwiggle=$alnwiggle --numthreads=$numthreads $inputbank.expels.fasta $inputbank.expels.bnk\".");
			#}
			unless (unlink("$inputbank.expels.fasta")) {
				&errorMessage(__LINE__, "Cannot delete \"$inputbank.expels.fasta\"");
			}
			unless (unlink("$inputbank.expels.fasta.qual")) {
				&errorMessage(__LINE__, "Cannot delete \"$inputbank.expels.fasta.qual\"");
			}
			# bank-combine
			if (system("$bankcombine $inputbank.temp.bnk $inputbank $inputbank.expels.bnk")) {
				&errorMessage(__LINE__, "Cannot run \"$bankcombine $inputbank.temp.bnk $inputbank $inputbank.expels.bnk\".");
			}
			&deleteBanks($inputbank, "$inputbank.expels.bnk");
			unless (rename("$inputbank.temp.bnk", $inputbank)) {
				&errorMessage(__LINE__, "Cannot rename \"$inputbank.temp.bnk\" to \"$inputbank\".");
			}
		}
	}
	if (-e "$inputbank.txt") {
		unless (unlink("$inputbank.txt")) {
			&errorMessage(__LINE__, "Cannot delete \"$inputbank.txt\"");
		}
	}
}

sub deleteBanks {
	foreach my $bankname (@_) {
		while (glob("$bankname/*")) {
			unless (unlink($_)) {
				&errorMessage(__LINE__, "Cannot delete \"$_\".");
			}
		}
		unless (rmdir("$bankname")) {
			&errorMessage(__LINE__, "Cannot delete \"$bankname\".");
		}
	}
}

sub cleanBank {
	my $bankname = shift(@_);
	my $keep = join('|', @_);
	while (glob("$bankname/*")) {
		if ($_ !~ /(?:$keep)\.[0-9a-z\.]+$/ && !unlink($_)) {
			&errorMessage(__LINE__, "Cannot delete \"$_\".");
		}
	}
}

# error message
sub errorMessage {
	my $lineno = shift(@_);
	my $message = shift(@_);
	print(STDERR "ERROR!: line $lineno\n$message\n");
	print(STDERR "If you want to read help message, run this script without options.\n");
	exit(1);
}

sub helpMessage {
	print(STDERR <<"_END");
Usage
=====
assams options input output

Command line options
====================
--assemblemode=EXACT|UNGAPPED|GAPPED|E+U|E+G|U+G|ALL
  Specify assemble mode. (default: E+G)

--minident=DECIMAL(,DECIMAL,DECIMAL)
  Specify minimum identity percentage. First value will be used in primary
assembly, second one will be used in secondary assembly, and last one
will be used in mapping. (default: 0.97,0.97,0.97)

--minovllen=INTEGER
  Specify minimum overlap length. (default: 100)

--maxnhits=INTEGER
  Specify maximum number of hits per read. 0 means unlimited.
(default: 10000)

--strand=PLUS|BOTH
  Specify search strand. (default: both)

--mergemode=ROUGH|DRAFT|NORMAL|ACCURATE
  Specify merging mode. (default: normal)

--alnwiggle=INTEGER
  Specify alignment wiggle. (default: 11)

--minnseq=INTEGER
  Specify minimum number of sequences of each splitted file.
(default: 1000)

--maxnsplit=INTEGER
  Specify maximum number of splitting. (default: unlimited)

--noeid
  If this option is specified, EIDs of contigs will not been
generated. (default: off)

-q, --qualfile=FILENAME
  Specify .qual file name. (default: inputfile.qual)

--numthreads=INTEGER
  Specify the number of threads. (default: 1)

Acceptable input file formats
=============================
FASTA (+.qual)
AMOS bank
_END
	exit;
}
